class Constants:
    JELLYFISH_USER_AGENT = f'jellyfish/1.0'
    # This is how many bytes are in a MB. This is used when
    # determining how much data to upload to S3 (we generally)
    # batch files by ~50 MBs
    MB_SIZE_IN_BYTES = 1048576
    JIRA_ISSUES_UNCOMPRESSED_FILE_SIZE = 50 * MB_SIZE_IN_BYTES
    # 250 is the default largest value we can pull from Jira Server
    # For cloud it's 100, but we have logic to check what Jira
    # limits us to and reduce our batch size to that
    MAX_ISSUE_API_BATCH_SIZE = 250
    # use the 10k issues pull new sync method
    NEW_SYNC_FF_NAME = 'makara-jf-ingest-use-new-sync-2024Q4'



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from types import MappingProxyType

from jf_ingest.config import JiraDownloadConfig
from jf_ingest.data_manifests.manifest_base import ManifestSource


# Some Jellyfish core dependencies were removed from this definition.
# project_keys_to_classification_type is provided by the caller instead of relying on
# ambient context or Django ORM. This should also make the class more testable and reusable.
@dataclass
class ManifestAdapter(ABC):
    '''
    Abstract class for getting different Jira Manifests
    '''

    company_slug: str
    config: JiraDownloadConfig
    manifest_source: ManifestSource
    excluded_project_keys: set[str] = field(default_factory=set())
    project_keys_to_classification_type: dict[str, str] = field(default_factory=MappingProxyType({}))

    def __init__(self,
                 config: JiraDownloadConfig,
                 manifest_source: ManifestSource,
                 project_keys_to_classification_type: dict[str, str] = MappingProxyType({})):
        self.company_slug = config.company_slug
        self.config = config
        self.manifest_source = manifest_source
        self.excluded_project_keys = set(config.exclude_projects)
        self.project_keys_to_classification_type = project_keys_to_classification_type

    @abstractmethod
    def get_users_count(self) -> int:
        pass

    @abstractmethod
    def get_fields_count(self) -> int:
        pass

    @abstractmethod
    def get_resolutions_count(self) -> int:
        pass

    @abstractmethod
    def get_issue_types_count(self) -> int:
        pass

    @abstractmethod
    def get_issue_link_types_count(self) -> int:
        pass

    @abstractmethod
    def get_priorities_count(self) -> int:
        pass

    @abstractmethod
    def get_project_data_dicts(self) -> list[dict]:
        pass

    @abstractmethod
    # Boards count is both a 'global' and a project_id
    # based value because you can have a personal Jira Board
    # that is associated with only a user and NOT a project.
    # This appears as the total board count being different
    # from the value you would get if you summed up the board_count
    # across all project manifests
    def get_boards_count(self) -> int:
        pass

    @abstractmethod
    def get_issues_count(self) -> int:
        pass

    @abstractmethod
    def get_project_versions_count(self) -> int:
        pass

    # EVERYTHING BELOW IS PROJECT SPECIFIC FUNCTIONS

    @abstractmethod
    def test_basic_auth_for_project(self, project_id: int) -> bool:
        pass

    @abstractmethod
    def get_project_versions_count_for_project(self, project_id: int) -> int:
        pass

    @abstractmethod
    def get_issues_count_for_project(self, project_id: int) -> int:
        pass

    @abstractmethod
    def get_issues_data_count_for_project(self, project_id: int) -> int:
        pass

    @abstractmethod
    def get_last_updated_for_project(self, project_id):
        pass



import logging

from jf_ingest import logging_helper

logger = logging.getLogger(__name__)


def _count_log_records_by_message(records):
    record_message_counts = {r.message: 0 for r in records}
    for record in records:
        record_message_counts[record.message] += 1

    return record_message_counts


def test_log_entry_exit_log_level(caplog):
    informational_log_message = "This is a test log"

    @logging_helper.log_entry_exit()
    def _test_function_for_wrapper():
        logging.info(informational_log_message)

    ########################################################################
    # TEST WITH LEVEL SET AT INFO (NO DEBUG MESSAGES SHOULD SHOW)
    ########################################################################
    caplog.set_level(level=logging.INFO)
    _test_function_for_wrapper()

    record_message_counts = _count_log_records_by_message(caplog.records)

    # Assert that the starting and ending log got called exactly once
    assert f"{_test_function_for_wrapper.__name__}: Starting" not in record_message_counts.keys()
    assert f"{_test_function_for_wrapper.__name__}: Ending" not in record_message_counts.keys()
    # Assert that the informational message appeared twice
    assert record_message_counts[informational_log_message] == 1
    assert len(caplog.records) == 1

    ########################################################################
    # TEST WITH LEVEL SET AT DEBUG
    ########################################################################
    # Reset caplog
    caplog.clear()
    caplog.set_level(level=logging.DEBUG)
    _test_function_for_wrapper()

    record_message_counts = _count_log_records_by_message(caplog.records)

    # Assert that the starting and ending log got called exactly once
    assert record_message_counts[f"{_test_function_for_wrapper.__name__}: Starting"] == 1
    assert record_message_counts[f"{_test_function_for_wrapper.__name__}: Ending"] == 1
    # Assert that the informational message appeared twice
    assert record_message_counts[informational_log_message] == 1
    assert len(caplog.records) == 3


def test_log_for_loop_iters_info_level(caplog):
    total_iters = 10
    log_every = 1

    ########################################################################
    # TEST WITH LEVEL SET AT INFO (NO DEBUG MESSAGES SHOULD SHOW)
    ########################################################################
    caplog.set_level(level=logging.INFO)
    logged_information_messages = []

    for i in range(10):
        with logging_helper.log_loop_iters("test_log_loop_iters", i, log_every):
            info_message = f"Iter {i}"
            logger.info(info_message)
            logged_information_messages.append(info_message)

    record_message_counts = _count_log_records_by_message(caplog.records)

    assert len(record_message_counts.keys()) == total_iters

    ########################################################################
    # TEST WITH LEVEL SET AT DEBUG
    ########################################################################
    caplog.clear()
    caplog.set_level(level=logging.DEBUG)

    for i in range(10):
        with logging_helper.log_loop_iters("test_log_loop_iters", i, log_every):
            info_message = f"Iter {i}"
            logger.info(info_message)
            logged_information_messages.append(info_message)

    record_message_counts = _count_log_records_by_message(caplog.records)

    # For every iter, we should have an additional 2
    assert len(record_message_counts.keys()) == (total_iters + (total_iters * 2))

    ########################################################################
    # TEST WITH LEVEL SET AT DEBUG (log every iter set to 2)
    ########################################################################
    log_every = 2
    caplog.clear()
    caplog.set_level(level=logging.DEBUG)

    for i in range(10):
        with logging_helper.log_loop_iters("test_log_loop_iters", i, log_every):
            info_message = f"Iter {i}"
            logger.info(info_message)
            logged_information_messages.append(info_message)

    record_message_counts = _count_log_records_by_message(caplog.records)

    # For every 2 iters, we should have an additional 2 debug logs
    assert len(record_message_counts.keys()) == (total_iters + (total_iters))


def test_log_standard_error_smoke_test(caplog):
    caplog.set_level(logging.WARNING)
    msg_args = ["TEST MESSAGE"]
    error_code = 0000
    logging_helper.log_standard_error(
        level=logging.WARNING, error_code=error_code, msg_args=msg_args
    )
    assert len(caplog.records) == 1
    assert caplog.records[0].message == logging_helper.generate_standard_error_msg(
        error_code=error_code, msg_args=msg_args
    )


def test_send_to_agent_log_basic(caplog):
    caplog.set_level(logging.DEBUG)
    info_log_message_for_agent_log = 'This is a normal INFO message for only the Agent log'
    info_log_message_for_not_agent_log = 'This is a normal INFO message not just for agent log'
    logging_helper.send_to_agent_log_file(info_log_message_for_agent_log)
    logger.info(info_log_message_for_not_agent_log)

    # Sanity check
    assert len(caplog.records) == 2
    # Assert that the log that is tagged as an agent log file log should be one
    assert caplog.records[0].message == info_log_message_for_agent_log
    assert caplog.records[0].__dict__[logging_helper.AGENT_LOG_TAG] == True
    assert caplog.records[0].levelno == logging.INFO  # Assert default level is INFO
    # Assert that the log that should not be tagged as an agent log file log is not tagged as one
    assert caplog.records[1].message == info_log_message_for_not_agent_log
    assert logging_helper.AGENT_LOG_TAG not in caplog.records[1].__dict__


def test_send_to_agent_log_with_varying_level(caplog):
    caplog.set_level(logging.DEBUG)
    logging_level = logging.ERROR
    info_log_message_for_agent_log = (
        f'This is a normal {logging_level} message for only the Agent log'
    )
    info_log_message_for_not_agent_log = (
        f'This is a normal {logging_level} message not just for agent log'
    )
    logging_helper.send_to_agent_log_file(info_log_message_for_agent_log, level=logging_level)
    logger.error(info_log_message_for_not_agent_log)

    # Sanity check
    assert len(caplog.records) == 2
    # Assert that the log that is tagged as an agent log file log should be one
    assert caplog.records[0].message == info_log_message_for_agent_log
    assert caplog.records[0].__dict__[logging_helper.AGENT_LOG_TAG] == True
    assert caplog.records[0].levelno == logging_level  # Assert we retain level
    # Assert that the log that should not be tagged as an agent log file log is not tagged as one
    assert caplog.records[1].message == info_log_message_for_not_agent_log
    assert logging_helper.AGENT_LOG_TAG not in caplog.records[1].__dict__


def test_send_to_agent_log_with_extra_data(caplog):
    caplog.set_level(logging.DEBUG)
    logging_level = logging.DEBUG
    extra_field_key = 'foo'
    extra_field_value = 'bar'
    extra_field_key_two = 10000
    extra_field_value_two = 1234
    extras = {extra_field_key: extra_field_value, extra_field_key_two: extra_field_value_two}
    info_log_message_for_agent_log = 'TEST LOG WITH EXTRAS'

    # EXECUTE FUNCTION
    logging_helper.send_to_agent_log_file(
        info_log_message_for_agent_log, level=logging_level, extra=extras
    )

    # Sanity check
    assert len(caplog.records) == 1
    agent_record = caplog.records[0]
    # Assert that the log that is tagged as an agent log file log should be one
    assert agent_record.message == info_log_message_for_agent_log
    assert agent_record.levelno == logging_level  # Assert we retain level
    # Assert extras are there
    log_extras = agent_record.__dict__
    assert (
        log_extras[logging_helper.AGENT_LOG_TAG] == True
    )  # ALWAYS RETAIN THAT THIS IS AN AGENT LOG RECORD
    # Assert we also have the extra extra data
    assert log_extras[extra_field_key] == extra_field_value
    assert log_extras[extra_field_key_two] == extra_field_value_two



import unittest
from unittest.mock import MagicMock, Mock

from jf_ingest.config import JiraDownloadConfig
from jf_ingest.data_manifests.jira.generator import generate_project_manifest, process_global_jira_data, ManifestAdapter


class TestGenerator(unittest.TestCase):
    def test_generate_project_manifest(self):
        # Arrange
        mock_create_jira_project_manifest = Mock()
        mock_reset_current_tenant = Mock()
        mock_manifest_adapter = MagicMock(spec=ManifestAdapter)
        mock_manifest_adapter.config = MagicMock(spec=JiraDownloadConfig)
        mock_manifest_adapter.manifest_source = 'test_source'
        mock_manifest_adapter.company_slug = 'test_company'
        mock_manifest_adapter.excluded_project_keys = []
        mock_manifest_adapter.get_issues_count_for_project.return_value = 10
        mock_manifest_adapter.get_project_versions_count_for_project.return_value = 5
        mock_manifest_adapter.get_last_updated_for_project.return_value = '2023-12-05'
        mock_manifest_adapter.project_keys_to_classification_type = {'test_project': 'test_classification'}
        mock_manifest_adapter.test_basic_auth_for_project.return_value = True

        project_data_dict = {'key': 'test_project', 'id': '1'}
        classifications_name_lookup = {'test_classification': 'Test Classification'}

        # Act
        result = generate_project_manifest(
            manifest_adapter=mock_manifest_adapter,
            project_data_dict=project_data_dict,
            classifications_name_lookup=classifications_name_lookup,
            create_jira_project_manifest=mock_create_jira_project_manifest,
            reset_current_tenant=mock_reset_current_tenant
        )

        # Assert
        mock_create_jira_project_manifest.assert_called_once_with(
            company='test_company',
            data_source='test_source',
            project_id='1',
            project_key='test_project',
            issues_count=10,
            version_count=5,
            excluded=False,
            pull_from=mock_manifest_adapter.config.earliest_issue_dt,
            last_issue_updated_date='2023-12-05',
            classification='test_classification',
            classification_str='Test Classification',
        )
        mock_reset_current_tenant.assert_called_once()
        self.assertEqual(result, mock_create_jira_project_manifest.return_value)

    def test_process_global_jira_data(self):
        # Arrange
        mock_create_jira_data_manifest = Mock()
        mock_reset_current_tenant = Mock()
        mock_manifest_adapter = MagicMock(spec=ManifestAdapter)
        mock_manifest_adapter.config = MagicMock(spec=JiraDownloadConfig)
        mock_manifest_adapter.company_slug = 'test_company'
        mock_manifest_adapter.manifest_source = 'test_source'
        mock_manifest_adapter.config.earliest_issue_dt = '2023-12-05'
        mock_manifest_adapter.get_users_count.return_value = 100
        mock_manifest_adapter.get_fields_count.return_value = 50
        mock_manifest_adapter.get_resolutions_count.return_value = 10
        mock_manifest_adapter.get_issue_types_count.return_value = 20
        mock_manifest_adapter.get_issue_link_types_count.return_value = 30
        mock_manifest_adapter.get_priorities_count.return_value = 40
        mock_manifest_adapter.get_project_data_dicts.return_value = [{'key': 'test_project', 'id': '1'}]
        mock_manifest_adapter.get_boards_count.return_value = 60
        mock_manifest_adapter.get_project_versions_count.return_value = 70
        mock_manifest_adapter.get_issues_count.return_value = 80

        # Act
        result = process_global_jira_data(
            manifest_adapter=mock_manifest_adapter,
            create_jira_data_manifest=mock_create_jira_data_manifest,
            reset_current_tenant=mock_reset_current_tenant
        )

        # Assert
        mock_create_jira_data_manifest.assert_called_once_with(
            company='test_company',
            data_source='test_source',
            pull_from='2023-12-05',
            users_count=100,
            fields_count=50,
            resolutions_count=10,
            issue_types_count=20,
            issue_link_types_count=30,
            priorities_count=40,
            projects_count=1,
            boards_count=60,
            project_versions_count=70,
            issues_count=80,
            project_manifests=[],
            encountered_errors_for_projects={},
        )
        mock_reset_current_tenant.assert_called_once()
        self.assertEqual(result, mock_create_jira_data_manifest.return_value)


from contextlib import contextmanager

import requests_mock

from jf_ingest.jf_jira.downloaders import download_fields
from tests.jf_jira.utils import _register_jira_uri_with_file, get_jira_mock_connection


@contextmanager
def _field_context_manager():
    with requests_mock.Mocker() as m:
        _register_jira_uri_with_file(m, "field", f"api_responses/fields.json")
        yield


def test_get_fields_once():
    expected_field = {
        "id": "statuscategorychangedate",
        "key": "statuscategorychangedate",
        "name": "Status Category Changed",
        "custom": False,
        "orderable": False,
        "navigable": True,
        "searchable": True,
        "clauseNames": ["statusCategoryChangedDate"],
        "schema": {"type": "datetime", "system": "statuscategorychangedate"},
    }

    with _field_context_manager():
        fields = download_fields(
            jira_connection=get_jira_mock_connection(),
            include_fields=[],
            exclude_fields=[],
        )

    assert len(fields) == 4
    assert fields[0] == expected_field
    for field in fields:
        assert "id" in field.keys()
        assert "key" in field.keys()
        assert "name" in field.keys()


def test_get_fields_exclude_list():
    with _field_context_manager():
        fields = download_fields(
            jira_connection=get_jira_mock_connection(),
            include_fields=[],
            exclude_fields=["customfield_10070"],
        )

    assert len(fields) == 3


def test_get_fields_include_list():
    with _field_context_manager():
        fields = download_fields(
            jira_connection=get_jira_mock_connection(),
            include_fields=["customfield_10070"],
            exclude_fields=[],
        )

    assert len(fields) == 1



import enum
import json
import logging
import tempfile
from collections import namedtuple
from dataclasses import dataclass, field
from datetime import datetime, timedelta, timezone
from typing import Any, TextIO, Union

import pytz
import requests
from dateutil import parser

from jf_ingest.constants import Constants
from jf_ingest.jf_git.exceptions import GitProviderUnavailable
from jf_ingest.jf_git.standardized_models import StandardizedRepository
from jf_ingest.utils import format_datetime_to_ingest_timestamp

logger = logging.getLogger(__name__)

JELLYFISH_API_BASE = "https://app.jellyfish.co"

# IssueListForDownload = namedtuple('IssueList', ['id_to_key_normal_updated', 'id_to_key_deleted', 'id_to_key_changed'])
# IssuesReturned = namedtuple('IssuesReturned', ['issues_list', 'total_issues', 'parent_ids'])
# IssueDownloadingResult = namedtuple('IssueDownloadingResult', ['downloaded_ids', 'discovered_parent_ids', 'total_batches'])


class JiraAuthMethod(enum.IntEnum):
    BasicAuth = 1
    AtlassianConnect = 2


class GitProvider(enum.Enum):
    # these come from an entry in the config file, so we want it to be something readable
    GITHUB = "GITHUB"


class GitProviderInJellyfishRepo(enum.Enum):
    # duplication of the class GitProvider values in the jellyfish repo so we can maintain consistency
    GITHUB = 'GH'


# Needs to subclass str to be serializable
class IngestionType(str, enum.Enum):
    AGENT = "AGENT"
    DIRECT_CONNECT = "DIRECT_CONNECT"


@dataclass
class IssueMetadata:
    id: str
    key: str
    updated: datetime
    project_id: str = (
        None  # NOTE: This field is optionally set, and generally only used for detected re-keys
    )
    # The following fields are used for detecting redownloads
    epic_link_field_issue_key: str = None
    parent_field_issue_key: str = None
    parent_id: str = None

    def __post_init__(self):
        """Post Init is called here to properly type everything by doing typecasts"""
        self.id = str(self.id)
        self.key = str(self.key)
        self.updated = (
            self.updated if isinstance(self.updated, datetime) else parser.parse(self.updated)
        )

        # Sanity recasts to make sure everything is a string
        if self.project_id:
            self.project_id = str(self.project_id)
        if self.epic_link_field_issue_key:
            self.epic_link_field_issue_key = str(self.epic_link_field_issue_key)
        if self.parent_field_issue_key:
            self.parent_field_issue_key = str(self.parent_field_issue_key)
        if self.parent_id:
            self.parent_id = str(self.parent_id)

    @staticmethod
    def deserialize_json_str_to_issue_metadata(
        json_str: str,
    ) -> Union["IssueMetadata", list["IssueMetadata"]]:
        """Helper function to deserialize a JSON type object to an IssueMetadata object, or a list of IssueMetadata Objects

        Raises:
            Exception: Raises an expection if the provided json str is not a valid JSON str

        Returns:
            _type_: Either a list of IssueMetadata Objects, or an IssueMetadata Object
        """
        json_data: Union[list, dict] = json.loads(json_str)
        if type(json_data) == list:
            list_of_issue_metadata = []
            for item in json_data:
                item['updated'] = datetime.fromisoformat(item['updated'])
                list_of_issue_metadata.append(IssueMetadata(**item))
            return list_of_issue_metadata
        elif type(json_data) == dict:
            json_data['updated'] = datetime.fromisoformat(json_data['updated'])
            return IssueMetadata(**json_data)
        else:
            raise Exception(
                f'Unrecognized type for deserialize_json_str_to_issue_metadata, type={type(json_data)}'
            )

    @staticmethod
    def from_json(
        json_source: Union[str, bytes, TextIO]
    ) -> Union["IssueMetadata", list["IssueMetadata"]]:
        """Generalized wrapper for the deserialize_json_str_to_issue_metadata function, which deserializes JSON Strings to Python objects

        Returns:
            _type_: Either a list of IssueMetadata objects or an IssueMetaData object, depending on what was supplied
        """
        if isinstance(json_source, str):
            return IssueMetadata.deserialize_json_str_to_issue_metadata(json_source)
        elif isinstance(json_source, bytes):
            return IssueMetadata.deserialize_json_str_to_issue_metadata(json_source.decode("utf-8"))
        elif isinstance(json_source, TextIO):
            json_str = json_source.read()
            return IssueMetadata.deserialize_json_str_to_issue_metadata(json_str)

    @staticmethod
    def to_json_str(issue_metadata: Union["IssueMetadata", list["IssueMetadata"]]) -> str:
        """This is a STATIC helper function that can serialize both a LIST of issue_metadata and a singluar
        IssueMetadata to a JSON string!

        Args:
            issue_metadata (Union[&quot;IssueMetadata&quot;, list[&quot;IssueMetadata&quot;]]): Either a list of IssueMetadata, or a singluar IssueMetadata object

        Returns:
            str: A serialized JSON str
        """

        def _serializer(value):
            if isinstance(value, datetime):
                return value.isoformat()
            elif isinstance(value, IssueMetadata):
                return value.__dict__
            else:
                return str(value)

        return json.dumps(issue_metadata, default=_serializer)

    @staticmethod
    def init_from_jira_issue(issue: dict, project_id: str = None, skip_parent_data: bool = False):
        fields: dict = issue.get("fields", {})
        return IssueMetadata(
            id=issue["id"],
            key=issue["key"],
            project_id=project_id,
            updated=parser.parse(fields.get("updated")) if fields.get("updated") else None,
            parent_id=fields.get("parent", {}).get("id") if not skip_parent_data else None,
            parent_field_issue_key=fields.get("parent", {}).get("key")
            if not skip_parent_data
            else None,
        )

    @staticmethod
    def init_from_jira_issues(issues=list[dict], skip_parent_data: bool = False):
        return [
            IssueMetadata.init_from_jira_issue(issue, skip_parent_data=skip_parent_data)
            for issue in issues
        ]

    # Define a hashing function so that we can find uniqueness
    # easily using sets
    def __hash__(self) -> str:
        return hash(self.id)

    def __eq__(self, __o) -> bool:
        return hash(self) == hash(__o)


@dataclass
class GitAuthConfig:
    # These fields are shared between all Git type authentication
    # schemes (I think)
    company_slug: str = None
    # Note: token is rarely provided, as we often do some type of
    # git JWT authentication to get a time sensitive login token.
    # See how we authenticate specifically for Github as an example:
    # GithubClient._get_app_access()
    token: str = None
    base_url: str = None
    verify: bool = True
    session: requests.Session = None


@dataclass
class GithubAuthConfig(GitAuthConfig):
    """This is specifically for Github Cloud and Github Enterprise!"""

    installation_id: str = (None,)
    app_id: str = (None,)
    private_key: str = (None,)

    def __post_init__(self):
        if self.installation_id and self.app_id and self.private_key:
            pass
        elif self.token:
            pass
        else:
            logger.warning(
                f'Github Auth Config requires either a provided Token or an installation_id, app_id, and private_key'
            )


@dataclass
class GitConfig:
    """This is the generalized Ingest Configuration for all Git Providers (Github, Gitlab, etc)

    Raises:
        GitProviderUnavailable: This error gets thrown in the git_provider is not currently supported

    Returns:
        GitConfig: A git config data object
    """

    company_slug: str
    instance_slug: str
    instance_file_key: str
    git_provider: GitProvider
    git_auth_config: GitAuthConfig
    use_agent: bool = False

    # If not provided we will use the default Github API
    url: str = None

    # Jellyfish Feature Flags
    jf_options: dict[str, Any] = field(default_factory=dict)

    # Maps when we should pull commits from for each repository (by ['org_login']['repo_id'] name)
    repos_to_skip_pull_prs_for: dict[str, datetime] = field(default_factory=dict)
    repos_to_prs_last_updated: dict[str, datetime] = field(default_factory=dict)
    # Maps the oldest PR/commit date we have in our system
    repos_to_commits_backpopulated_to: dict[str, datetime] = field(default_factory=dict)
    repos_to_prs_backpopulated_to: dict[str, datetime] = field(default_factory=dict)

    # Git Organization info
    # In Github these are called Organizations.
    # In Gitlab these are called Groups.
    discover_orgs: bool = False
    git_organizations: list[str] = field(default_factory=list)
    pull_from: datetime = datetime.min.replace(tzinfo=pytz.UTC)

    # D.C. only logic
    skip_pulling_users: bool = False
    skip_pulling_repos: bool = False
    skip_discover_organizations: bool = False
    repos_in_jellyfish: list[StandardizedRepository] = field(default_factory=list)
    # Mark whether we should pull all commits AND branches for each repo.
    # In agent, this should be True. In Jellyfish, it is controlled by the JiraInstance
    pull_all_commits_and_branches: bool = False
    # In jellyfish this is controlled at the instance level and at the repo level
    repo_id_to_pull_all_commits_and_branches: dict[str, bool] = field(default_factory=dict)
    force_full_backpopulation_pull: bool = False
    backpopulation_window_days: int = 60
    # A list of Repo IDs that are marked as quiescent in the Jellyfish DB
    quiescent_repos: set[str] = field(default_factory=set)

    # Agent specific configuration
    excluded_organizations: list[str] = field(default_factory=list)
    included_repos: list[str] = field(default_factory=list)
    excluded_repos: list[str] = field(default_factory=list)
    included_branches_by_repo: dict[str, str] = field(default_factory=dict)
    git_redact_names_and_urls: bool = False
    git_strip_text_content: bool = False

    def __post_init__(self):
        if type(self.git_provider) == str:
            if self.git_provider.upper() in [
                GitProvider.GITHUB.value,
                GitProviderInJellyfishRepo.GITHUB.value,
            ]:
                self.git_provider = GitProviderInJellyfishRepo.GITHUB
            else:
                raise GitProviderUnavailable(
                    f'Provided git provider is not available: {self.git_provider}'
                )

    def get_pull_from_for_commits(
        self,
    ) -> datetime:
        """This is a helper function for getting the 'pull_from' date for Commits for any given repository.

        The normal, daily case -- we pull commits based on a commit date greater than some date in the past.  A commit
        with a commit date of 1 week ago could be present in a branch today even though it wasn't yesterday (if, say, it
        was just merged into the branch today) -- so we don't want earliest_commit_date_to_pull_if_not_backpop to be too
        small.  30 days ago is probably reasonable.

        Returns:
            datetime: A pull from date to pull commits from (in UTC). Typically about 60 days in the past
        """
        return datetime.now(tz=timezone.utc) - timedelta(days=31)

    def get_pull_from_for_prs(
        self,
        repo_id: str,
    ) -> datetime:
        """This is a helper function for getting the specified 'pull_from' date for PRs, for a given repo.
        This value is generally the date of the latest commit we have in our Jellyfish database. If a datetime
        is not set for this repo ID, we will default to the value set at pull_from

        Args:
            org_login (str): The org login for the parent Git Organization
            repo_id (str): The ID for the repository we are concerned with

        Returns:
            datetime: A pull from date to pull PRs from (in UTC)
        """
        return self.repos_to_prs_last_updated.get(repo_id, self.pull_from)

    def get_backpopulated_date_for_commits(
        self,
        repo_id: int,
    ) -> datetime:
        """This is a helper function to get the oldest commit date that we have in our system. It is needed
        to determine if we need to run a backpopulation algorithm

        Args:
            org_login (str): The org login for the parent Git Organization
            repo_id (str): The ID for the repository we are concerned with

        Returns:
            datetime: A pull from date to pull commits from (in UTC)
        """
        return self.repos_to_commits_backpopulated_to.get(repo_id, None)

    def get_backpopulated_date_for_prs(
        self,
        repo_id: str,
    ) -> datetime:
        """This is a helper function to get the oldest PR date that we have in our system. It is needed
        to determine if we need to run a backpopulation algorithm

        Args:
            org_login (str): The org login for the parent Git Organization
            repo_id (str): The ID for the repository we are concerned with

        Returns:
            datetime: A pull from date to pull PRs from (in UTC)
        """
        return self.repos_to_prs_backpopulated_to.get(repo_id, None)


@dataclass
class JiraAuthConfig:
    # Provides an authenticated connection to Jira, without the additional
    # settings needed for download/ingest.
    company_slug: str = None
    url: str = None
    # NOTE: Used in the User-Agent header. Not related
    # to the Jellyfish Agent
    user_agent: str = Constants.JELLYFISH_USER_AGENT
    # Used for Basic Auth
    user: str = None
    password: str = None
    personal_access_token: str = None
    # Used for Atlassian Direct Connect
    jwt_attributes: dict[str, str] = field(default_factory=dict)
    bypass_ssl_verification: bool = False
    required_email_domains: bool = False
    is_email_required: bool = False
    available_auth_methods: list[JiraAuthMethod] = field(default_factory=list)
    connect_app_active: bool = False


@dataclass
class JiraDownloadConfig(JiraAuthConfig):
    # Indicates whether the email augmentation should be performed. This is needed
    # when testing connect-driven data refreshes locally since the connect app used
    # for testing does not (and can not) have the required "Email" permission. As
    # such, this allows us to skip that step when necessary.
    should_augment_emails: bool = True

    # Jira Server Information
    gdpr_active: bool = None

    include_fields: list[str] = None
    exclude_fields: list[str] = None
    # User information
    force_search_users_by_letter: bool = False
    search_users_by_letter_email_domain: str = None

    # Projects information
    include_projects: list[str] = None
    exclude_projects: list[str] = None
    include_project_categories: list[str] = None
    exclude_project_categories: list[str] = None

    # Boards/Sprints
    download_sprints: bool = False

    # Issues
    skip_issues: bool = False
    only_issues: bool = False
    full_redownload: bool = False
    earliest_issue_dt: datetime = datetime.min
    issue_batch_size: int = Constants.MAX_ISSUE_API_BATCH_SIZE
    issue_download_concurrent_threads: int = 1
    recursively_download_parents: bool = False  # When provided, we will download parents of parents, and then those parents, until we find all parents
    # Dict of Issue ID (str) to IssueMetadata Object
    jellyfish_issue_metadata: list[IssueMetadata] = None
    jellyfish_issue_ids_for_redownload: set[str] = None
    jellyfish_project_ids_to_keys: dict = None

    # worklogs
    download_worklogs: bool = False
    # Potentially solidify this with the issues date, or pull from
    work_logs_pull_from: datetime = datetime.min

    # Jira Ingest Feature Flags
    feature_flags: dict = field(default_factory=dict)

    # Agent
    uses_agent: bool = False


@dataclass
class IngestionConfig:
    # upload info
    save_locally: bool = True
    upload_to_s3: bool = False
    local_file_path: str = None
    company_slug: str = None
    timestamp: str = None

    # Jira Auth Info and Download Configuration
    jira_config: JiraDownloadConfig = None

    # Git Config data
    # NOTE: Unlike Jira, we can have multiple Git Configurations.
    # Each Git Configuration maps to a Git Instance
    git_configs: list[GitConfig] = field(default_factory=list)

    # JF specific config
    jellyfish_api_token: str = None
    jellyfish_api_base: str = JELLYFISH_API_BASE
    ingest_type: IngestionType = None

    def __post_init__(self):
        if not self.timestamp:
            self.timestamp = format_datetime_to_ingest_timestamp(datetime.utcnow())

        if not self.local_file_path:
            self.local_file_path = f"{tempfile.TemporaryDirectory().name}/{self.timestamp}"


@dataclass
class IssueDownloadingResult:
    downloaded_ids: set[str]
    discovered_parent_ids: set[str]
    total_batches: int


@dataclass
class IssueListForDownload:
    id_to_key_updated: dict[str, str]
    id_to_key_deleted: dict[str, str]
    id_to_key_changed: dict[str, str]


@dataclass
class IssuesReturnedFromRemote:
    issues_list: list[dict]
    parent_ids: set[str]
    total_issues: int



from datetime import datetime, timedelta, timezone

from jf_ingest.config import GitConfig
from jf_ingest.jf_git.adapters import (
    BackpopulationWindow,
    determine_commit_backpopulation_window,
    determine_pr_backpopulation_window,
)
from jf_ingest.jf_git.standardized_models import StandardizedOrganization, StandardizedRepository

GIT_CONFIG_DEFAULTS = {
    'company_slug': 'test_company',
    'instance_slug': 'test_instance',
    'instance_file_key': 'test_FILEKEY',
    'git_provider': 'GITHUB',
    'git_auth_config': None,
}

GIT_TEST_ORG = StandardizedOrganization(id='test_org', name='Test Org', login='Test-Org', url='')
GIT_TEST_REPO = StandardizedRepository(
    id='1',
    name='test_repo',
    full_name='test_repo_full_name',
    url='',
    is_fork=False,
    default_branch_name='main',
    default_branch_sha='test',
    organization=GIT_TEST_ORG,
)


def create_git_config(*args, **kwargs):
    return GitConfig(*args, **kwargs)


def test_backpopulation_window_helper_confirm_always_none_cases():
    pull_from = datetime(2024, 1, 1)
    object_last_pulled_date = pull_from

    git_config_args = {
        'pull_from': pull_from,
        'force_full_backpopulation_pull': True,
        'backpopulation_window_days': 10,
        'repos_to_prs_backpopulated_to': {GIT_TEST_REPO.id: object_last_pulled_date},
        'repos_to_commits_backpopulated_to': {GIT_TEST_REPO.id: object_last_pulled_date},
        **GIT_CONFIG_DEFAULTS,
    }
    config = create_git_config(**git_config_args)
    repo = GIT_TEST_REPO

    prs_back_population_window = determine_pr_backpopulation_window(config=config, repo=repo)
    commits_back_population_window = determine_commit_backpopulation_window(
        config=config, repo=repo
    )
    assert commits_back_population_window == None
    assert prs_back_population_window == None

    ######################################################
    git_config_args = {
        'pull_from': pull_from,
        'force_full_backpopulation_pull': True,
        'backpopulation_window_days': 10,
        'repos_to_prs_backpopulated_to': {
            GIT_TEST_REPO.id: object_last_pulled_date - timedelta(days=1)
        },
        'repos_to_commits_backpopulated_to': {
            GIT_TEST_REPO.id: object_last_pulled_date - timedelta(days=1)
        },
        **GIT_CONFIG_DEFAULTS,
    }
    config = create_git_config(**git_config_args)
    repo = GIT_TEST_REPO

    prs_back_population_window = determine_pr_backpopulation_window(config=config, repo=repo)
    commits_back_population_window = determine_commit_backpopulation_window(
        config=config, repo=repo
    )
    assert commits_back_population_window == None
    assert prs_back_population_window == None


def test_backpopulation_window_helper_with_force_pull():
    pull_from = datetime(2024, 1, 1)
    object_last_pulled_date = datetime(2024, 2, 1)
    git_config_args = {
        'pull_from': pull_from,
        'repos_to_prs_backpopulated_to': {GIT_TEST_REPO.id: object_last_pulled_date},
        'force_full_backpopulation_pull': True,
        **GIT_CONFIG_DEFAULTS,
    }
    config = create_git_config(**git_config_args)
    repo = GIT_TEST_REPO

    prs_back_population_window = determine_pr_backpopulation_window(config=config, repo=repo)
    commits_back_population_window = determine_commit_backpopulation_window(
        config=config, repo=repo
    )
    assert prs_back_population_window == BackpopulationWindow(pull_from, object_last_pulled_date)

    #############################
    git_config_args = {
        'pull_from': pull_from,
        'repos_to_commits_backpopulated_to': {GIT_TEST_REPO.id: object_last_pulled_date},
        'force_full_backpopulation_pull': True,
        **GIT_CONFIG_DEFAULTS,
    }
    config = create_git_config(**git_config_args)
    repo = GIT_TEST_REPO

    prs_back_population_window = determine_pr_backpopulation_window(config=config, repo=repo)
    commits_back_population_window = determine_commit_backpopulation_window(
        config=config, repo=repo
    )
    assert commits_back_population_window == BackpopulationWindow(
        pull_from, object_last_pulled_date
    )


def test_backpopulation_window_helper_with_custom_days_window():
    pull_from = datetime(2024, 1, 1, tzinfo=timezone.utc)
    object_last_pulled_date = datetime(2024, 2, 1, tzinfo=timezone.utc)
    backpop_window_days = 10
    target_backpopulation_start_date = object_last_pulled_date - timedelta(days=backpop_window_days)
    git_config_args = {
        'pull_from': pull_from,
        'repos_to_prs_backpopulated_to': {GIT_TEST_REPO.id: object_last_pulled_date},
        'backpopulation_window_days': backpop_window_days,
        **GIT_CONFIG_DEFAULTS,
    }
    config = create_git_config(**git_config_args)
    repo = GIT_TEST_REPO

    prs_back_population_window = determine_pr_backpopulation_window(config=config, repo=repo)
    commits_back_population_window = determine_commit_backpopulation_window(
        config=config, repo=repo
    )
    assert prs_back_population_window == BackpopulationWindow(
        target_backpopulation_start_date, object_last_pulled_date
    )

    #############################
    git_config_args = {
        'pull_from': pull_from,
        'repos_to_commits_backpopulated_to': {GIT_TEST_REPO.id: object_last_pulled_date},
        'backpopulation_window_days': backpop_window_days,
        **GIT_CONFIG_DEFAULTS,
    }
    config = create_git_config(**git_config_args)
    repo = GIT_TEST_REPO

    prs_back_population_window = determine_pr_backpopulation_window(config=config, repo=repo)
    commits_back_population_window = determine_commit_backpopulation_window(
        config=config, repo=repo
    )
    assert commits_back_population_window == BackpopulationWindow(
        target_backpopulation_start_date, object_last_pulled_date
    )

    ##############################
    git_config_args = {
        'pull_from': pull_from,
        'repos_to_prs_backpopulated_to': {GIT_TEST_REPO.id: object_last_pulled_date},
        'repos_to_commits_backpopulated_to': {GIT_TEST_REPO.id: object_last_pulled_date},
        'backpopulation_window_days': backpop_window_days,
        **GIT_CONFIG_DEFAULTS,
    }
    config = create_git_config(**git_config_args)
    repo = GIT_TEST_REPO

    prs_back_population_window = determine_pr_backpopulation_window(config=config, repo=repo)
    commits_back_population_window = determine_commit_backpopulation_window(
        config=config, repo=repo
    )
    assert commits_back_population_window == BackpopulationWindow(
        target_backpopulation_start_date, object_last_pulled_date
    )
    assert prs_back_population_window == BackpopulationWindow(
        target_backpopulation_start_date, object_last_pulled_date
    )


def test_backpopulation_window_when_there_is_no_data_in_jellyfish():
    pull_from = datetime(2024, 1, 1, tzinfo=timezone.utc)
    start_time = datetime.now().astimezone(timezone.utc)
    backpop_window_days = 10
    git_config_args = {
        'pull_from': pull_from,
        'backpopulation_window_days': backpop_window_days,
        **GIT_CONFIG_DEFAULTS,
    }
    config = create_git_config(**git_config_args)
    repo = GIT_TEST_REPO

    prs_back_population_window = determine_pr_backpopulation_window(config=config, repo=repo)
    commits_back_population_window = determine_commit_backpopulation_window(
        config=config, repo=repo
    )
    assert commits_back_population_window.backpopulation_window_end >= start_time
    assert prs_back_population_window.backpopulation_window_end >= start_time
    assert (
        prs_back_population_window.backpopulation_window_end
        - prs_back_population_window.backpopulation_window_start
    ).days == 10
    assert (
        commits_back_population_window.backpopulation_window_end
        - commits_back_population_window.backpopulation_window_start
    ).days == 10

    #####################################################
    git_config_args = {
        'pull_from': pull_from,
        'force_full_backpopulation_pull': True,
        **GIT_CONFIG_DEFAULTS,
    }
    config = create_git_config(**git_config_args)
    repo = GIT_TEST_REPO

    prs_back_population_window = determine_pr_backpopulation_window(config=config, repo=repo)
    commits_back_population_window = determine_commit_backpopulation_window(
        config=config, repo=repo
    )
    assert commits_back_population_window.backpopulation_window_end >= start_time
    assert prs_back_population_window.backpopulation_window_end >= start_time
    assert prs_back_population_window.backpopulation_window_start == pull_from
    assert commits_back_population_window.backpopulation_window_start == pull_from



import json
import logging
import time
from datetime import datetime, timezone
from typing import Generator, Optional

import jwt
import requests
from dateutil import parser
from requests.utils import default_user_agent

from jf_ingest.config import GithubAuthConfig
from jf_ingest.constants import Constants
from jf_ingest.graphql_utils import (
    GQL_PAGE_INFO_BLOCK,
    GQLRateLimit,
    GqlRateLimitedExceptionInner,
    get_gql_rate_limit,
)
from jf_ingest.jf_git.exceptions import GitAuthenticationException, GqlRateLimitExceededException
from jf_ingest.utils import retry_for_status, retry_session

logger = logging.getLogger(__name__)

GIT_DEFAULT_API_ENDPOINT = 'https://api.github.com'


def parse_date(dt):
    if dt is None:
        return None
    return parser.parse(dt)


class GithubClient:
    # This uses GQL to hit the Github API!

    GITHUB_GQL_USER_FRAGMENT = "... on User {login, id: databaseId, email, name, url}"
    # Need to make a second, special actor fragment to make sure we grab
    # the proper ID from either a bot or a User
    GITHUB_GQL_ACTOR_FRAGMENT = """
        ... on Actor 
            { 
                login 
                ... on User { id: databaseId, email, name } 
                ... on Bot { id: databaseId}
            }
    """
    # NOTE: On the author block here, we have a type GitActor
    # We cannot always get the email from the nested user object,
    # so pull whatever email we can from the gitActor top level object.
    # (we can't get the email from the user object bc of variable privacy configuration)
    GITHUB_GQL_COMMIT_FRAGMENT = f"""
        ... on Commit {{
            sha: oid
            url
            author {{
                ... on GitActor {{
                    email
                    name
                    user {{ id: databaseId, login }}
                }}
            }}
            message
            committedDate
            authoredDate
            parents {{totalCount}}
        }}
    """
    GITHUB_GQL_SHORT_REPO_FRAGMENT = "... on Repository {name, id:databaseId, url}"

    # The PR query is HUGE, we shouldn't query more than about 25 at a time
    MAX_PAGE_SIZE_FOR_PR_QUERY = 25

    def __init__(
        self,
        github_auth_config: GithubAuthConfig,
        **kwargs,
    ):
        """This is a wrapper class for interacting with the Github GQL endpoint. It supports
        both cloud and enterprise. It works by either accepting a provided token or by using
        the other args to set up JWT authentication (see: https://developer.github.com/apps/building-github-apps/authenticating-with-github-apps/)

        Args:
            token (str, optional): An optional authentication token. If not provided we will use the other args to get a JWT token. See GithubClient._get_app_access_token. Defaults to None.
            installation_id (str, optional): A client provided installation ID, used for JWT authentication. Defaults to None.
            app_id (str, optional): A client provided App Id, used for JWT authentication. Defaults to None.
            private_key (str, optional): A user provided private key, used for JWT authentication. Defaults to None.
            base_url (str, optional): An optional base url, used by Github Enterprise customers. Defaults to None.
            verify (bool, optional): Session arg. Defaults to True.
            session (requests.Session, optional): An optional session to pass this client, if you've already set up an authenticated session. Defaults to None.
            company_slug (str, optional): An optional str giving this class context what company this client is for. Defaults to None.
        """

        self.company_slug = github_auth_config.company_slug
        self.gql_base_url = self.get_github_gql_base_url(base_url=github_auth_config.base_url)
        # We need to hit the REST API for some API calls, see get_organization_by_name and _get_app_access_token
        self.rest_api_url = github_auth_config.base_url or GIT_DEFAULT_API_ENDPOINT

        if github_auth_config.token is None:
            self.installation_id = github_auth_config.installation_id
            self.app_id = github_auth_config.app_id
            self.private_key = github_auth_config.private_key
            self.token, self.token_expiration = self._get_app_access_token()
            self.uses_jwt = True
        else:
            self.token = github_auth_config.token
            self.token_expiration = None
            self.uses_jwt = False

        self.session = github_auth_config.session or retry_session(**kwargs)
        self.session.verify = True
        self.session.headers.update(
            {
                'Authorization': f'token {self.token}',
                'Accept': 'application/vnd.github+json',
                'User-Agent': f'{Constants.JELLYFISH_USER_AGENT} ({default_user_agent()})',
            }
        )

    @staticmethod
    def datetime_to_gql_str_format(_datetime: datetime) -> str:
        return _datetime.replace(tzinfo=timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")

    @staticmethod
    def get_github_gql_base_url(base_url: str) -> str:
        # the URL for GQL things is different from what we need to get a token
        if base_url and 'api/v3' in base_url:
            # Github server clients provide an API with a trailing '/api/v3'
            # replace this with the graphql endpoint
            return base_url.replace('api/v3', 'api/graphql')
        else:
            return f'{base_url or GIT_DEFAULT_API_ENDPOINT}/graphql'

    @staticmethod
    def create_jwt_token(app_id, private_key):
        now = int(time.time()) - 60
        payload = {"iat": now, "exp": now + 600, "iss": app_id}
        jwt_token = jwt.encode(payload, key=private_key, algorithm="RS256")
        jwt_token = jwt_token.decode('utf-8')
        return jwt_token

    def get_raw_result(self, query_body: str, max_attempts: int = 7) -> dict:
        """Gets the raw results from a Graphql Query.

        Args:
            query_body (str): A query body to hit GQL with
            max_attempts (int, optional): The number of retries we should make when we specifically run into GQL Rate limiting. This value is important if the GQL endpoint doesn't give us (or gives us a malformed) rate limit header. Defaults to 7.

        Raises:
            GqlRateLimitExceededException: A custom exception if we run into GQL rate limiting and we run out of attempts (based on max_attempts)
            Exception: Any other random exception we encounter, although the big rate limiting use cases are generally covered

        Returns:
            dict: A raw dictionary result from GQL
        """
        attempt_number = 1
        while True:
            try:
                response: requests.Response = retry_for_status(
                    self.session.post, url=self.gql_base_url, json={'query': query_body}
                )
                response.raise_for_status()
                json_str = response.content.decode()
                json_data = json.loads(json_str)
                if 'errors' in json_data:
                    if len(json_data['errors']) == 1:
                        error_dict: dict = json_data['errors'][0]
                        if error_dict.get('type') == 'RATE_LIMITED':
                            raise GqlRateLimitedExceptionInner(
                                error_dict.get('message', 'Rate Limit hit in GQL')
                            )
                    raise Exception(
                        f'Exception encountered when trying to query: {query_body}. Error: {json_data["errors"]}'
                    )
                return json_data
            except requests.exceptions.HTTPError as e:
                # Our GQL connection times out after 60 minutes, if we encounter a 401
                # attempt to re-establish a connection
                if e.response.status_code == 401:
                    self._update_token()
                    continue
                # We can get transient 403 level errors that have to do with rate limiting,
                # but aren't directly related to the above GqlRateLimitedExceptionInner logic.
                # Do a simple retry loop here
                elif e.response.status_code == 403:
                    pass
                else:
                    raise

                # Raise if we've passed our limit
                if attempt_number > max_attempts:
                    raise

                sleep_time = attempt_number**2
                # Overwrite sleep time if github gives us a specific wait time
                if e.response.headers.get('retry-after') and attempt_number == 1:
                    retry_after = int(e.response.headers.get('retry-after'))
                    if retry_after > (60 * 5):
                        # if the given wait time is more than 5 minutes, call their bluff
                        # and try the experimental backoff approach
                        pass
                    elif retry_after <= 0:
                        # if the given wait time is negative ignore their suggestion
                        pass
                    else:
                        # Add three seconds for gracetime
                        sleep_time = retry_after + 3

                logger.warning(
                    f'A secondary rate limit was hit. Sleeping for {sleep_time} seconds. (attempt {attempt_number}/{max_attempts})',
                )
                time.sleep(sleep_time)
            except GqlRateLimitedExceptionInner:
                if attempt_number > max_attempts:
                    raise GqlRateLimitExceededException(
                        f'Exceeded maximum retry limit ({max_attempts})'
                    )

                rate_limit_info: GQLRateLimit = get_gql_rate_limit(
                    session=self.session, base_url=self.gql_base_url
                )
                reset_at: datetime = rate_limit_info.reset_at
                reset_at_timestamp = reset_at.timestamp()
                curr_timestamp = datetime.utcnow().timestamp()

                sleep_time = reset_at_timestamp - curr_timestamp

                # Sometimes github gives a reset time way in the
                # future. But rate limits reset each hour, so don't
                # wait longer than that
                sleep_time = min(sleep_time, 3600)

                # Sometimes github gives a reset time in the
                # past. In that case, wait for 5 mins just in case.
                if sleep_time <= 0:
                    sleep_time = 300
                logger.warning(
                    f'GQL Rate Limit hit. Sleeping for {sleep_time} seconds',
                )
                time.sleep(sleep_time)
            finally:
                attempt_number += 1

    def page_results_gql(
        self, query_body: str, path_to_page_info: str, cursor: str = 'null'
    ) -> Generator[dict, None, None]:
        """This is a helper function for paging results from GraphQL. It expects
        a query body to hit Graphql with that has a %s marker after the "after:"
        key word, so that we can inject a cursor into the query. This will allow
        us to page results in GraphQL.
        To use this function properly, the section you are trying to page MUST
        INCLUDE VALID PAGE INFO (including the hasNext and endCursor attributes)

        Args:
            query_body (str): The query body to hit GraphQL with
            path_to_page_info (str): A string of period separated words that lead
            to the part of the query that we are trying to page. Example: data.organization.userQuery
            cursor (str, optional): LEAVE AS NULL - this argument is use recursively to page. The cursor
            will continuously go up, based on the endCursor attribute in the GQL call. Defaults to 'null'.

        Yields:
            Generator[dict, None, None]: This function yields each item from all the pages paged, item by item
        """
        if not cursor:
            cursor = 'null'
        hasNextPage = True
        while hasNextPage:
            # Fetch results
            result = self.get_raw_result(query_body=(query_body % cursor))

            yield result

            # Get relevant data and yield it
            path_tokens = path_to_page_info.split('.')
            for token in path_tokens:
                result = result[token]

            page_info = result['pageInfo']
            # Need to grab the cursor and wrap it in quotes
            _cursor = page_info['endCursor']
            # If endCursor returns null (None), break out of loop
            hasNextPage = page_info['hasNextPage'] and _cursor
            cursor = f'"{_cursor}"'

    def _get_app_access_token(self):
        """
        Authenticating a github app requires encoding the installation id, and private key
        into a JWT token in order to request a access_token from the server.
        See: https://developer.github.com/apps/building-github-apps/authenticating-with-github-apps/
        """

        base_url = self.rest_api_url

        # create the jwt_token to authenticate request
        jwt_token = GithubClient.create_jwt_token(app_id=self.app_id, private_key=self.private_key)

        # fetch token from the server
        response = requests.post(
            url=f'{base_url}/app/installations/{self.installation_id}/access_tokens',
            headers={
                'Accept': 'application/vnd.github.machine-man-preview+json',
                'Authorization': f'Bearer {jwt_token}',
                'User-Agent': f'{Constants.JELLYFISH_USER_AGENT} ({default_user_agent()})',
            },
        )
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if e.response.status_code in (403, 404):
                msg = (
                    f"Got HTTP {e.response.status_code} when attempting to create a GithubClient the this address: {base_url}. "
                    "This means our app does not have permission to access the customer's github instance. "
                    "This usually happens when either permissions have intentionally revoked or if an access token expires. "
                    "Set the JFGithubInstance to enabled=False to skip this GitHub instance but allow pulling from any other instances to complete successfully. "
                )
                logger.error(msg)
                raise GitAuthenticationException(msg, original_exception=e)

            raise

        response_data = response.json()
        token = response_data['token']
        token_expiration = parse_date(response_data['expires_at'])
        logger.info(
            f'Obtained token successfully - new token will expire in 60m at {token_expiration}'
        )
        return token, token_expiration

    def _update_token(self):
        # refresh token
        self.token, self.token_expiration = self._get_app_access_token()
        self.session.headers.update(
            {
                'Accept': 'application/json',
                'User-Agent': f'{Constants.JELLYFISH_USER_AGENT} ({default_user_agent()})',
                'Authorization': f'token {self.token}',
            }
        )

    def _check_token_expiration(self):
        if self.token_expiration:
            mins_until_expiration = (
                self.token_expiration - datetime.now(timezone.utc)
            ).total_seconds() / 60
            if mins_until_expiration < 10:
                logger.info(
                    f'Token is going to expire in {mins_until_expiration:.1f} minutes -- obtaining a new token.'
                )
                self._update_token()

    # This is for commits, specifically the 'author' block within them.
    # On the GQL side of things, these are specifically a distinct type of object,
    # GitActor. It has a nested user object, but the quality of data within it
    # is variable due to a users privacy settings. Email, for example, is often
    # not present in the child user block, so we always grab it from the top level.
    @staticmethod
    def _process_git_actor_gql_object(author: dict) -> dict:
        user: dict = author.get('user') or {}
        return {
            'id': user.get('id'),
            'login': user.get('login'),
            'email': author['email'],
            'name': author['name'],
        }

    def get_scopes_of_api_token(self):
        # Make an empty call against the orgs API to be quick
        # and get the OAuth scopes
        url = f'{self.rest_api_url}/orgs/'
        result = self.session.get(url)
        return result.headers.get('X-OAuth-Scopes')

    # This is for commits, specifically the 'author' block within them.
    # On the GQL side of things, these are specifically a distinct type of object,
    # GitActor. It has a nested user object, but the quality of data within it
    # is variable due to a users privacy settings. Email, for example, is often
    # not present in the child user block, so we always grab it from the top level.
    @staticmethod
    def _process_git_actor_gql_object(author: dict) -> dict:
        user: dict = author.get('user') or {}
        return {
            'id': user.get('id'),
            'login': user.get('login'),
            'email': author['email'],
            'name': author['name'],
        }

    # HACK: This call will actually use the REST endpoint
    # Agent clients are supposed to have the [org:read] scope,
    # but many of them don't. This wasn't a problem before
    # because the REST org API doesn't actually hide behind any perms...
    # TODO: Once we straighten out everybody's permissions we can sunset
    # this function
    def get_organization_by_login(self, login: str):
        # NOTE: We are hitting a different base url here!
        url = f'{self.rest_api_url}/orgs/{login}'

        result = self.session.get(url)

        # HACK: This appears to happen after we have been
        # rate-limited when hitting certain URLs, there is
        # likely a more elegant way to solve this but it takes
        # about an hour to test each time and it works.
        # NOTE: This is unlikely for this call, because we are only hitting
        # the org once for an entire git config
        if result.status_code == 403:
            result = self.session.get(url)

        result.raise_for_status()
        return result.json()

    def get_users(self, login: str) -> Generator[dict, None, None]:
        query_body = f"""{{
            organization(login: \"{login}\") {{
                userQuery: membersWithRole(first: 100, after: %s) {{
                    {GQL_PAGE_INFO_BLOCK}
                    users: nodes {{
                        {self.GITHUB_GQL_USER_FRAGMENT}
                    }}
                }}
            }}
        }}
        """
        for page in self.page_results_gql(
            query_body=query_body, path_to_page_info='data.organization.userQuery'
        ):
            for user in page['data']['organization']['userQuery']['users']:
                yield user

    def get_team_members(self, login: str, team_slug: str) -> Generator[dict, None, None]:
        query_body = f"""{{
            organization(login: \"{login}\") {{
                team(slug: \"{team_slug}\") {{
                    membersQuery: members(first: 100, after: %s) {{
                        {GQL_PAGE_INFO_BLOCK}
                        members: nodes {{
                            {self.GITHUB_GQL_USER_FRAGMENT}
                        }}
                    }}
                }}
            }}
        }}
        """
        for page in self.page_results_gql(
            query_body=query_body, path_to_page_info='data.organization.team.membersQuery'
        ):
            for member in page['data']['organization']['team']['membersQuery']['members']:
                yield member

    def get_teams(self, login: str) -> Generator[dict, None, None]:
        query_body = f"""{{
            organization(login: \"{login}\") {{
                teamsQuery: teams(first: 100, after: %s) {{
                    {GQL_PAGE_INFO_BLOCK}
                    teams: nodes {{
                        id
                        slug
                        name
                        description
                        membersQuery: members(first: 100) {{
                            {GQL_PAGE_INFO_BLOCK}
                            members: nodes {{
                                {self.GITHUB_GQL_USER_FRAGMENT}
                            }}
                        }}
                    }}
                }}
            }}
        }}
        """
        for page in self.page_results_gql(
            query_body=query_body, path_to_page_info='data.organization.teamsQuery'
        ):
            for team in page['data']['organization']['teamsQuery']['teams']:
                if team['membersQuery']['pageInfo']['hasNextPage']:
                    logger.debug(
                        f'Team {team["name"]} was detected as having more than {len(team["membersQuery"]["members"])} members, we need to page for additional members'
                    )
                    team['members'] = [
                        member
                        for member in self.get_team_members(login=login, team_slug=team['slug'])
                    ]
                else:
                    team['members'] = team['membersQuery']['members']

                yield team

    def get_repos(
        self, login: str, repo_filters: list[filter] = None
    ) -> Generator[dict, None, None]:
        query_body = f"""{{
            organization(login: "{login}") {{
                repoQuery: repositories(first: 50, after: %s) {{
                    {GQL_PAGE_INFO_BLOCK}
                    repos: nodes {{
                        ... on Repository {{
                            id: databaseId
                            name
                            fullName: nameWithOwner
                            url
                            isFork
                            defaultBranch: defaultBranchRef {{ name, target {{ sha: oid }} }}
                        }}
                    }}
                }}
            }}
        }}
        """
        for page in self.page_results_gql(
            query_body=query_body, path_to_page_info='data.organization.repoQuery'
        ):
            for api_repo in page['data']['organization']['repoQuery']['repos']:
                # Skip over excluded or ignore non-included
                if not all(filt(api_repo) for filt in repo_filters):
                    continue
                else:
                    yield api_repo

    def get_branches(self, login: str, repo_name: str) -> Generator[dict, None, None]:
        query_body = f"""{{
            organization(login: "{login}") {{
                repository(name: "{repo_name}") {{
                    ... on Repository {{
                        branchQuery: refs(refPrefix:"refs/heads/", first: 100, after: %s) {{
                            {GQL_PAGE_INFO_BLOCK}
                            branches: nodes {{
                                ... on Ref {{
                                    name
                                    target {{sha: oid}}
                                }}
                            }}
                        }}
                    }}
                }}
            }}
        }}
        """
        for page in self.page_results_gql(
            query_body=query_body, path_to_page_info='data.organization.repository.branchQuery'
        ):
            for api_branch in page['data']['organization']['repository']['branchQuery']['branches']:
                yield api_branch

    def get_commits_count(
        self, login: str, repo_name: str, branch_name: str, since: datetime
    ) -> int:
        query_body = f"""{{
            organization(login: "{login}") {{
                repo: repository(name: "{repo_name}") {{
                    ... on Repository {{
                        branch: ref(qualifiedName: "{branch_name}") {{
                            target {{
                                ... on Commit {{
                                    history(first: 0, since: "{self.datetime_to_gql_str_format(since)}") {{
                                        totalCount
                                    }}
                                }}
                            }}
                        }}
                    }}
                }}
            }}
        }}
        """
        return self.get_raw_result(query_body)['data']['organization']['repo']['branch']['target'][
            'history'
        ]['totalCount']

    def get_commits(
        self,
        login: str,
        repo_name: str,
        branch_name: str,
        since: datetime,
        until: Optional[datetime] = None,
    ) -> Generator[dict, None, None]:
        optional_until_clause = (
            f' until: "{self.datetime_to_gql_str_format(until)}",' if until else ""
        )
        query_body = f"""{{
            organization(login: "{login}") {{
                repo: repository(name: "{repo_name}") {{
                    ... on Repository {{
                        branch: ref(qualifiedName: "{branch_name}") {{
                            target {{
                                ... on Commit {{
                                    history(first: 100, since: "{self.datetime_to_gql_str_format(since)}", {optional_until_clause} after: %s) {{
                                        {GQL_PAGE_INFO_BLOCK}
                                        commits: nodes {{
                                            {self.GITHUB_GQL_COMMIT_FRAGMENT}
                                        }}
                                    }}
                                }}
                            }}
                        }}
                    }}
                }}
            }}
        }}
        """
        for page in self.page_results_gql(
            query_body=query_body, path_to_page_info='data.organization.repo.branch.target.history'
        ):
            for api_commit in page['data']['organization']['repo']['branch']['target']['history'][
                'commits'
            ]:
                # Overwrite Author block for backwards compatibility
                api_commit['author'] = self._process_git_actor_gql_object(api_commit['author'])
                yield api_commit

    #
    # PR Queries are HUGE, so pull out reusable blocks (comments, reviews, commits, etc)
    #
    def _get_pr_comments_query_block(self, enable_paging: bool = False):
        return f"""
            commentsQuery: comments(first: 100{', after: %s' if enable_paging else ''}) {{
                {GQL_PAGE_INFO_BLOCK}
                
                comments: nodes {{
                    author {{
                        {self.GITHUB_GQL_ACTOR_FRAGMENT}
                    }}
                    body
                    createdAt
                }}
            }}
        """

    # NOTE: There are comments associated with reviews that we need to fetch as well
    def _get_pr_reviews_query_block(self, enable_paging: bool = False):
        return f"""
            reviewsQuery: reviews(first: 25{', after: %s' if enable_paging else ''}) {{
                {GQL_PAGE_INFO_BLOCK}
                
                reviews: nodes {{
                    ... on PullRequestReview {{
                        author {{
                            {self.GITHUB_GQL_ACTOR_FRAGMENT}
                        }}
                        id: databaseId
                        state
                        # NOTE! We are paging for comments here as well!
                        {self._get_pr_comments_query_block()}
                    }}
                }}
            }}
        """

    def _get_pr_commits_query_block(self, enable_paging: bool = False):
        return f"""
            commitsQuery: commits(first: 50{', after: %s' if enable_paging else ''}) {{
                {GQL_PAGE_INFO_BLOCK}
                
                commits: nodes {{
                    ... on PullRequestCommit {{
                        commit {{
                            {self.GITHUB_GQL_COMMIT_FRAGMENT}
                        }}
                    }}
                }}
            }}
        """

    def get_prs_count(self, login: str, repo_name: str) -> int:
        query_body = f"""{{
            organization(login: "{login}") {{
                repo: repository(name: "{repo_name}") {{
                    prQuery: pullRequests(first: 1, orderBy: {{direction: DESC, field: UPDATED_AT}}) {{
                        totalCount
                    }}
                }}
            }}
        }}
        """
        return self.get_raw_result(query_body=query_body)['data']['organization']['repo'][
            'prQuery'
        ]['totalCount']

    def get_prs_metadata(self, login: str, repo_name: str) -> Generator[dict, None, None]:
        """
        Helper function, intended to be SUPER lightweight query against PRs. It returns PR Id (number) and the last updated date,
        as well as the current GQL cursor of the returned PR.

        Returned Format:
        {
            'cursor': 'gql_cursor_value',
            'pr': {
                'number': 1,
                'updatedAt': 'gql_formatted_date'
            }
        }
        """
        query_body = f"""{{
            organization(login: "{login}") {{
                repo: repository(name: "{repo_name}") {{
                    prQuery: pullRequests(first: 100, orderBy: {{direction: DESC, field: UPDATED_AT}}, after: %s) {{
                        {GQL_PAGE_INFO_BLOCK}
                        prs: edges {{ cursor, pr: node {{number, updatedAt}} }}
                    }}
                }}
            }}
        }}
        """
        for page in self.page_results_gql(
            query_body=query_body, path_to_page_info='data.organization.repo.prQuery'
        ):
            for api_pr in page['data']['organization']['repo']['prQuery']['prs']:
                yield api_pr

    # PR query is HUGE, see above GITHUB_GQL_PR_* blocks for reused code
    # page_size is optimally variable. Most repos only have a 0 to a few PRs day to day,
    # so sometimes the optimal page_size is 0. Generally, we should never go over 25
    def get_prs(
        self,
        login: str,
        repo_name: str,
        include_top_level_comments: bool = False,
        page_size: int = MAX_PAGE_SIZE_FOR_PR_QUERY,
        start_cursor: any = None,
    ) -> Generator[dict, None, None]:
        query_body = f"""{{
            organization(login: "{login}") {{
                repo: repository(name: "{repo_name}") {{
                    prQuery: pullRequests(first: {page_size}, orderBy: {{direction: DESC, field: UPDATED_AT}}, after: %s) {{
                        {GQL_PAGE_INFO_BLOCK}
                        prs: nodes {{
                            ... on PullRequest {{
                                id: number
                                number
                                additions
                                deletions
                                changedFiles
                                state
                                merged
                                createdAt
                                updatedAt
                                mergedAt
                                closedAt
                                title
                                body
                                url
                                baseRefName
                                headRefName
                                baseRepository {{ {self.GITHUB_GQL_SHORT_REPO_FRAGMENT} }}
                                headRepository {{ {self.GITHUB_GQL_SHORT_REPO_FRAGMENT} }}
                                author {{
                                    {self.GITHUB_GQL_ACTOR_FRAGMENT}
                                }}
                                mergedBy {{
                                    {self.GITHUB_GQL_ACTOR_FRAGMENT}
                                }}
                                mergeCommit {{
                                    {self.GITHUB_GQL_COMMIT_FRAGMENT}
                                }}
                                {self._get_pr_comments_query_block(enable_paging=False) if include_top_level_comments else ''}
                                {self._get_pr_reviews_query_block(enable_paging=False)}
                                {self._get_pr_commits_query_block(enable_paging=False)}
                            }}
                        }}
                    }}
                }}
            }}
        }}
        """
        transformed_start_cursor = f'"{start_cursor}"' if start_cursor else None
        for page in self.page_results_gql(
            query_body=query_body,
            path_to_page_info='data.organization.repo.prQuery',
            cursor=transformed_start_cursor,
        ):
            for api_pr in page['data']['organization']['repo']['prQuery']['prs']:
                # Process and add related PR data (comments, reviews, commits)
                # This may require additional API calls
                pr_number = api_pr['number']

                # Load reviews first because we use them in both reviews and comments
                reviews = (
                    [r for r in self.get_pr_reviews(login, repo_name, pr_number=pr_number)]
                    if api_pr['reviewsQuery']['pageInfo']['hasNextPage']
                    else api_pr['reviewsQuery']['reviews']
                )

                # NOTE: COMMENTS ARE WEIRD! They exist in there own API endpoint (these
                # are typically top level comments in a PR, considered an IssueComment)
                # but there are also comments associated with each review (typically only one)
                # The baseline for what we care about is the Review Level comment, pulled from
                # the reviews endpoint. Grabbing Top Level Comments is an optional feature flag

                # Grab the comments pulled from reviews. We ALWAYS want these!
                api_pr['comments'] = [
                    comment for review in reviews for comment in review['commentsQuery']['comments']
                ]

                # Grab the potentially optional top level comments
                if include_top_level_comments:
                    top_level_comments = (
                        [
                            comment
                            for comment in self.get_pr_top_level_comments(
                                login, repo_name, pr_number=pr_number
                            )
                        ]
                        if api_pr['commentsQuery']['pageInfo']['hasNextPage']
                        else api_pr['commentsQuery']['comments']
                    )
                    api_pr['comments'].extend(top_level_comments)

                api_pr['reviews'] = reviews

                api_pr['commits'] = (
                    [
                        commit
                        for commit in self.get_pr_commits(login, repo_name, pr_number=pr_number)
                    ]
                    if api_pr['commitsQuery']['pageInfo']['hasNextPage']
                    else [commit['commit'] for commit in api_pr['commitsQuery']['commits']]
                )

                # Do some extra processing on commits to clean up their weird author block
                for commit in api_pr['commits']:
                    commit['author'] = self._process_git_actor_gql_object(commit['author'])

                if api_pr['mergeCommit'] and api_pr['mergeCommit']['author']:
                    api_pr['mergeCommit']['author'] = self._process_git_actor_gql_object(
                        api_pr['mergeCommit']['author']
                    )

                yield api_pr

    def get_pr_top_level_comments(
        self, login: str, repo_name: str, pr_number: int
    ) -> Generator[dict, None, None]:
        query_body = f"""{{
            organization(login: "{login}") {{
                repo: repository(name: "{repo_name}") {{
                    pr: pullRequest(number: {pr_number}) {{
                        ... on PullRequest {{
                            {self._get_pr_comments_query_block(enable_paging=True)}
                        }}
                    }}
                }}
            }}
        }}
        """
        for page in self.page_results_gql(
            query_body=query_body, path_to_page_info='data.organization.repo.pr.commentsQuery'
        ):
            for api_pr_comment in page['data']['organization']['repo']['pr']['commentsQuery'][
                'comments'
            ]:
                yield api_pr_comment

    def get_pr_reviews(
        self, login: str, repo_name: str, pr_number: int
    ) -> Generator[dict, None, None]:
        query_body = f"""{{
            organization(login: "{login}") {{
                repo: repository(name: "{repo_name}") {{
                    pr: pullRequest(number: {pr_number}) {{
                        ... on PullRequest {{
                            {self._get_pr_reviews_query_block(enable_paging=True)}
                        }}
                    }}
                }}
            }}
        }}
        """
        for page in self.page_results_gql(
            query_body=query_body, path_to_page_info='data.organization.repo.pr.reviewsQuery'
        ):
            for api_pr_review in page['data']['organization']['repo']['pr']['reviewsQuery'][
                'reviews'
            ]:
                yield api_pr_review

    def get_pr_commits(
        self, login: str, repo_name: str, pr_number: int
    ) -> Generator[dict, None, None]:
        query_body = f"""{{
            organization(login: "{login}") {{
                repo: repository(name: "{repo_name}") {{
                    pr: pullRequest(number: {pr_number}) {{
                        ... on PullRequest {{
                            {self._get_pr_commits_query_block(enable_paging=True)}
                        }}
                    }}
                }}
            }}
        }}
        """
        for page in self.page_results_gql(
            query_body=query_body, path_to_page_info='data.organization.repo.pr.commitsQuery'
        ):
            for api_pr_commit in page['data']['organization']['repo']['pr']['commitsQuery'][
                'commits'
            ]:
                # Commit blocks are nested within the 'commits' block
                commit = api_pr_commit['commit']
                commit['author'] = self._process_git_actor_gql_object(commit['author'])
                yield commit

    def get_users_count(self, login: str) -> int:
        query_body = f"""{{
            organization(login: "{login}"){{
                    users: membersWithRole {{
                        totalCount
                    }}
                }}
            }}
        """
        # TODO: Maybe serialize the return results so that we don't have to do this crazy nested grabbing?
        return self.get_raw_result(query_body=query_body)['data']['organization']['users'][
            'totalCount'
        ]

    def get_repos_count(self, login: str) -> int:
        query_body = f"""{{
            organization(login: "{login}"){{
                    repos: repositories {{
                        totalCount
                    }}
                }}
            }}
        """
        # TODO: Maybe serialize the return results so that we don't have to do this crazy nested grabbing?
        return self.get_raw_result(query_body=query_body)['data']['organization']['repos'][
            'totalCount'
        ]

    def get_repo_manifest_data(
        self, login: str, page_size: int = 10
    ) -> Generator[dict, None, None]:
        query_body = f"""{{
            organization(login: "{login}") {{
                    repositories(first: {page_size}, after: %s) {{
                        pageInfo {{
                            endCursor
                            hasNextPage
                            
                        }}
                        repos: nodes {{
                            id: databaseId
                            name
                            url
                            defaultBranch: defaultBranchRef {{
                                name
                                target {{
                                    ... on Commit {{
                                        history {{
                                            totalCount
                                        }}
                                    }}
                                }}
                            }}
                            users: assignableUsers{{
                                totalCount
                            }}
                            prs: pullRequests {{
                                totalCount
                            }}
                            branches: refs(refPrefix:"refs/heads/") {{
                                totalCount
                            }}
                        }}
                    }}
                }}
            }}
        """
        path_to_page_info = 'data.organization.repositories'
        for result in self.page_results_gql(
            query_body=query_body, path_to_page_info=path_to_page_info
        ):
            for repo in result['data']['organization']['repositories']['repos']:
                yield repo

    def get_pr_manifest_data(
        self, login: str, repo_name: str, page_size=100
    ) -> Generator[dict, None, None]:
        query_body = f"""{{
                organization(login: "{login}") {{
                        repository(name: "{repo_name}") {{
                            name
                            id: databaseId
                            prs_query: pullRequests(first: {page_size}, after: %s) {{
                                pageInfo {{
                                    endCursor
                                    hasNextPage
                                }}
                                totalCount
                                prs: nodes {{
                                    updatedAt
                                    id: databaseId
                                    title
                                    number
                                    repository {{id: databaseId, name}}
                                }}
                            }}
                        }}
                    }}
                }}
        """

        path_to_page_info = 'data.organization.repository.prs_query'
        for result in self.page_results_gql(
            query_body=query_body, path_to_page_info=path_to_page_info
        ):
            for pr in result['data']['organization']['repository']['prs_query']['prs']:
                yield pr



import json
import tempfile

from jf_ingest import diagnostics

_test_diagnostic_1 = {"test": "1"}
_test_diagnostic_2 = {"test": 2}
_test_diagnostic_3 = {"test": "three"}


def test_open_diagnostics_file():
    with tempfile.TemporaryDirectory() as outdir:
        # Test opening and closing diagnostics file
        diagnostics.open_file(outdir=outdir)
        assert diagnostics._DIAGNOSTICS_FILE.name == f"{outdir}/diagnostics.json"
        diagnostics.close_file()
        assert diagnostics._DIAGNOSTICS_FILE == None

        # Test writing to a closed diagnostics file
        diagnostics._write_diagnostic(_test_diagnostic_1)
        assert diagnostics._DIAGNOSTICS_FILE == None

        # Test opening file twice (nothing should happen)
        diagnostics.open_file(outdir=outdir)
        diagnostics.open_file(outdir=outdir)
        assert diagnostics._DIAGNOSTICS_FILE != None
        assert diagnostics._DIAGNOSTICS_FILE.name == f"{outdir}/diagnostics.json"
        diagnostics._write_diagnostic(_test_diagnostic_1)
        diagnostics.close_file()
        assert diagnostics._DIAGNOSTICS_FILE == None


def test_write_diagnostic():
    with tempfile.TemporaryDirectory() as outdir:
        # Test opening and writing to diagnostics
        diagnostics.open_file(outdir=outdir)
        diagnostics._write_diagnostic(_test_diagnostic_1)
        diagnostics._write_diagnostic(_test_diagnostic_2)

        # TEST THAT WE CAN CLOSE AND REOPEN THE SAME FILE, AND APPEND DATA
        diagnostics.close_file()
        diagnostics.open_file(outdir=outdir)
        diagnostics._write_diagnostic(_test_diagnostic_3)
        diagnostics.close_file()

        with open(f"{outdir}/diagnostics.json", "r") as f:
            diagnostics_file_lines = f.readlines()
            assert json.loads(diagnostics_file_lines[0]) == _test_diagnostic_1
            assert json.loads(diagnostics_file_lines[1]) == _test_diagnostic_2
            assert json.loads(diagnostics_file_lines[2]) == _test_diagnostic_3



import logging
from enum import Enum
from typing import Generator, Optional

from jira import JIRA

from jf_ingest import diagnostics, logging_helper
from jf_ingest.config import (
    IngestionConfig,
    IngestionType,
    IssueDownloadingResult,
    IssueListForDownload,
    IssueMetadata,
    JiraAuthMethod,
    JiraDownloadConfig,
)
from jf_ingest.constants import Constants
from jf_ingest.file_operations import IngestIOHelper, SubDirectory
from jf_ingest.jf_jira.auth import get_jira_connection
from jf_ingest.jf_jira.downloaders import (
    detect_issues_needing_re_download,
    download_all_issue_metadata,
    download_boards_and_sprints,
    download_fields,
    download_issuelinktypes,
    download_issues,
    download_issues_from_new_sync,
    download_issuetypes,
    download_priorities,
    download_projects_and_versions_and_components,
    download_resolutions,
    download_statuses,
    download_users,
    download_worklogs,
    get_ids_from_difference_of_issue_metadata,
    get_issue_list_to_download,
    get_jira_search_batch_size,
    get_out_of_date_issue_ids,
)
from jf_ingest.utils import batch_iterable_by_bytes_size

logger = logging.getLogger(__name__)


class JiraObject(Enum):
    JiraFields = "jira_fields"
    JiraProjectsAndVersions = "jira_projects_and_versions"
    JiraUsers = "jira_users"
    JiraResolutions = "jira_resolutions"
    JiraIssueTypes = "jira_issuetypes"
    JiraLinkTypes = "jira_linktypes"
    JiraPriorities = "jira_priorities"
    JiraBoards = "jira_boards"
    JiraSprints = "jira_sprints"
    JiraBoardSprintLinks = "jira_board_sprint_links"
    JiraIssues = "jira_issues"
    JiraIssuesIdsDownloaded = "jira_issue_ids_downloaded"
    JiraIssuesIdsDeleted = "jira_issue_ids_deleted"
    JiraWorklogs = "jira_worklogs"
    JiraStatuses = "jira_statuses"


def load_issues_in_batches(
    issues_to_download: Generator[dict, None, None],
    ingest_io_helper: IngestIOHelper,
    ingest_config: IngestionConfig,
    batch_number_start: int = 0,
) -> IssueDownloadingResult:
    """given a generator object for issues to download; save them to disk in batches according to maximum size

    Args:
        issues_to_download: Generator object for issues to download
        ingest_io_helper: IngestIOHelper object for doing file operations
        ingest_config (IngestionConfig): A dataclass that holds several different configuration args for this task
        batch_number_start: starting index for the batch number (jira_issuesN.json). Used if you call this function multiple times

    Return:
        IssueDownloadingResult: A dataclass that holds the results of the download;
            - the ids for jira issues retrieved,
            - jira_ids for parents of those retrieved issues,
            - and the total number of batches it took
    """
    total_issue_batches = 0
    all_downloaded_issue_ids = set()
    collected_parent_ids = set()

    for batch_number, batch_issues in enumerate(
        batch_iterable_by_bytes_size(
            issues_to_download, batch_byte_size=Constants.JIRA_ISSUES_UNCOMPRESSED_FILE_SIZE
        ),
        start=batch_number_start,
    ):
        logging_helper.send_to_agent_log_file(
            f'Saving {len(batch_issues)} issues as batch number {batch_number}',
        )
        all_downloaded_issue_ids.update(set(issue['id'] for issue in batch_issues))
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraIssues.value,
            subdirectory=SubDirectory.JIRA,
            json_data=batch_issues,
            batch_number=batch_number,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )
        total_issue_batches += 1

        # we need to download parents of issues we just fetched
        collected_parent_ids.update(
            set(
                issue['fields']['parent']['id']
                for issue in batch_issues
                if 'parent' in issue['fields'].keys()
            )
        )

    logger.info(
        f"Successfully saved {len(all_downloaded_issue_ids)} Jira Issues in "
        f"{total_issue_batches} separate batches, with each batch limited to "
        f"{round(Constants.JIRA_ISSUES_UNCOMPRESSED_FILE_SIZE / Constants.MB_SIZE_IN_BYTES)}MB per batch"
    )

    return IssueDownloadingResult(
        discovered_parent_ids=collected_parent_ids,
        downloaded_ids=all_downloaded_issue_ids,
        total_batches=total_issue_batches,
    )


def pull_issues_old_metadata_path(
    jira_connect_or_fallback_connection: JIRA,
    jira_issues_batch_size: int,
    jira_config: JiraDownloadConfig,
    projects_and_versions: list[dict],
    ingest_io_helper: IngestIOHelper,
    ingest_config: IngestionConfig,
):
    """ "Default" old path for pulling issues
        get all metadata from jira, compare with metadata from JF, decide what to pull based on that
    Args:
        jira_connect_or_fallback_connection: either the basic connection or the special atlassian connect keys
        jira_issues_batch_size: batch size to use for actually downloading issues (much lower than the 10k for IDs)
        jira_config: jira config for jf_ingest from JF
        projects_and_versions: list of projects and versions to pull issues from
        ingest_io_helper: IngestIOHelper object for doing file operations
        ingest_config: full jf_ingest config
    """

    logging_helper.send_to_agent_log_file("Running 100 batched sync path")
    logging_helper.send_to_agent_log_file(
        f"All issue operations will use a batch_size of {jira_issues_batch_size}",
    )

    #######################################################################
    # Pull Issue Metadata
    #######################################################################
    issue_metadata_from_jira: list[IssueMetadata] = download_all_issue_metadata(
        jira_connection=jira_connect_or_fallback_connection,
        project_keys=[
            proj["key"]
            for proj in projects_and_versions
            if proj["key"] not in jira_config.exclude_projects
        ],
        earliest_issue_dt=jira_config.earliest_issue_dt,
        num_parallel_threads=jira_config.issue_download_concurrent_threads,
        batch_size=jira_issues_batch_size,
        recursively_download_parents=jira_config.recursively_download_parents,
    )

    issues_generator = download_issues(
        jira_connection=jira_connect_or_fallback_connection,
        full_redownload=jira_config.full_redownload,
        issue_download_concurrent_threads=jira_config.issue_download_concurrent_threads,
        jira_issues_batch_size=jira_issues_batch_size,
        issue_metadata_from_jellyfish=jira_config.jellyfish_issue_metadata,
        issue_metadata_from_jira=issue_metadata_from_jira,
        include_fields=jira_config.include_fields,
        exclude_fields=jira_config.exclude_fields,
    )

    all_downloaded_issue_ids = []
    total_issue_batches = 0

    for batch_number, batch_issues in enumerate(
        batch_iterable_by_bytes_size(
            issues_generator, batch_byte_size=Constants.JIRA_ISSUES_UNCOMPRESSED_FILE_SIZE
        )
    ):
        logging_helper.send_to_agent_log_file(
            f'Saving {len(batch_issues)} issues as batch number {batch_number}',
        )
        all_downloaded_issue_ids.extend([issue['id'] for issue in batch_issues])
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraIssues.value,
            subdirectory=SubDirectory.JIRA,
            json_data=batch_issues,
            batch_number=batch_number,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )
        total_issue_batches += 1

    logger.info(
        f"Successfully saved {len(all_downloaded_issue_ids)} Jira Issues in {total_issue_batches} separate batches, "
        f"with each batch limited to "
        f"{round(Constants.JIRA_ISSUES_UNCOMPRESSED_FILE_SIZE / Constants.MB_SIZE_IN_BYTES)}MB per batch"
    )

    # Write Issues
    ingest_io_helper.write_json_to_local_or_s3(
        object_name=JiraObject.JiraIssuesIdsDownloaded.value,
        json_data=[int(issue_id) for issue_id in all_downloaded_issue_ids],
        subdirectory=SubDirectory.JIRA,
        save_locally=ingest_config.save_locally,
        upload_to_s3=ingest_config.upload_to_s3,
    )

    deleted_issue_ids = get_ids_from_difference_of_issue_metadata(
        jira_config.jellyfish_issue_metadata, issue_metadata_from_jira
    )

    return all_downloaded_issue_ids, deleted_issue_ids


def pull_issues_new_sync_path(
    jira_connect_or_fallback_connection: JIRA,
    jira_config: JiraDownloadConfig,
    projects_and_versions: list[dict],
    jira_issues_batch_size: int,
    ingest_io_helper: IngestIOHelper,
    ingest_config: IngestionConfig,
):
    """Pulls issues using the "new" sync path, based on our ability to pull 10,000 jira issue IDs at one time
        as long as you only ask for `id` and `key` in the returned fields. Because we get no more fields we detect
        deletes as the absence of an issue on remote which we have local, and key changes as id match key mismatch
    Args:
        jira_connect_or_fallback_connection: either the basic connection or the special atlassian connect keys
        jira_config: jira config for jf_ingest from JF
        projects_and_versions: list of projects and versions to pull issues from
        jira_issues_batch_size: batch size to use for actually downloading issues (much lower than the 10k for IDs)
        ingest_io_helper: IngestIOHelper object for doing file operations
        ingest_config: full jf_ingest config

    """

    logging_helper.send_to_agent_log_file("Running 10k batched sync path")
    # get list of issues to download from remote
    if type(jira_config.jellyfish_issue_ids_for_redownload) is str:
        issues_for_redownload = [
            i.strip("[]").strip() for i in jira_config.jellyfish_issue_ids_for_redownload.split(",")
        ]
    else:
        issues_for_redownload = [str(i) for i in jira_config.jellyfish_issue_ids_for_redownload]
    issues_to_download: IssueListForDownload = get_issue_list_to_download(
        jira_connection=jira_connect_or_fallback_connection,
        jellyfish_issue_metadata=jira_config.jellyfish_issue_metadata,
        project_key_to_id={
            proj["key"]: proj["id"]
            for proj in projects_and_versions
            if proj["key"] not in jira_config.exclude_projects
        },
        earliest_issue_dt=jira_config.earliest_issue_dt,
        full_redownload=jira_config.full_redownload,
        issue_ids_for_redownload=issues_for_redownload,
    )

    issues_generator = download_issues_from_new_sync(
        jira_connection=jira_connect_or_fallback_connection,
        issue_download_concurrent_threads=jira_config.issue_download_concurrent_threads,
        jira_issues_batch_size=jira_issues_batch_size,
        issues_to_download=issues_to_download,
        include_fields=jira_config.include_fields,
        exclude_fields=jira_config.exclude_fields,
    )
    download_results: IssueDownloadingResult = load_issues_in_batches(
        issues_generator, ingest_io_helper, ingest_config
    )
    # don't need to download parents of issues that already were fetched
    to_be_downloaded_parent_ids = (
        download_results.discovered_parent_ids - download_results.downloaded_ids
    )

    all_downloaded_ids = download_results.downloaded_ids
    file_batch_number = download_results.total_batches

    # fetch parents
    # NOTE: For the general Jellyfish use case, we only need to go "one level deep" on parents,
    # i.e. pull the set of missing parents but NOT the parents of those parents (the "grandparents").
    # There is an optional feature, however, that will allow us to pull the parents of parents of parents
    # until we've pulled all parents. This is NOT the general use case, and will likely add a lot of computation
    # time to this function, so we advise you only use it in very specific use cases
    depth_level = 1
    while to_be_downloaded_parent_ids:
        logger.info(
            f'Attempting to pull more parents ({len(to_be_downloaded_parent_ids)} detected parents to pull). Some of them are {sorted(list(to_be_downloaded_parent_ids))[:30]} Depth level: {depth_level}'
        )
        issues_generator = download_issues_from_new_sync(
            jira_connection=jira_connect_or_fallback_connection,
            issue_download_concurrent_threads=jira_config.issue_download_concurrent_threads,
            jira_issues_batch_size=jira_issues_batch_size,
            issues_to_download=IssueListForDownload(to_be_downloaded_parent_ids, [], []),
            include_fields=jira_config.include_fields,
            exclude_fields=jira_config.exclude_fields,
        )
        parents_batch_result: IssueDownloadingResult = load_issues_in_batches(
            issues_to_download=issues_generator,
            ingest_io_helper=ingest_io_helper,
            ingest_config=ingest_config,
            batch_number_start=file_batch_number,
        )

        all_downloaded_ids.update(parents_batch_result.downloaded_ids)
        more_parent_ids = parents_batch_result.discovered_parent_ids - all_downloaded_ids

        if jira_config.recursively_download_parents:
            # TODO: We could probably reduce the numbers of files we upload by grouping
            # all extra parents in joined files, but I don't think it's a big deal to upload
            # a large number of small files
            file_batch_number += parents_batch_result.total_batches
            to_be_downloaded_parent_ids = more_parent_ids
            depth_level += 1
        else:
            logging_helper.send_to_agent_log_file(
                f'recursively_download_parents is set to {jira_config.recursively_download_parents}. Exiting execution'
            )
            break

    # Write Issues
    ingest_io_helper.write_json_to_local_or_s3(
        object_name=JiraObject.JiraIssuesIdsDownloaded.value,
        json_data=[int(issue_id) for issue_id in all_downloaded_ids],
        subdirectory=SubDirectory.JIRA,
        save_locally=ingest_config.save_locally,
        upload_to_s3=ingest_config.upload_to_s3,
    )

    return all_downloaded_ids, issues_to_download.id_to_key_deleted


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def load_and_push_jira_to_s3(ingest_config: IngestionConfig):
    """Loads data from JIRA, Dumps it to disk, and then uploads that data to S3

    All configuration for this object is done via the JIRAIngestionConfig object

    Args:
        config (IngestionConfig): A dataclass that holds several different configuration args for this task

    Returns:
        bool: Returns True on Success
    """

    # For Agent, we leverage the DEBUG log level for some things that are really INFO level. We do this
    # because we want to keep the output console in Agent as clean as possible, and we don't want to
    # interrupt the TQDM progress bar
    #
    # When we run this code in Prefect, however, we lose the DEBUG level logs because Prefect logging is
    # pretty rigid and hard to set with fine granularity. To get around this problem, use logger.log with
    # a variable log level
    logging_helper.set_ingestion_type(ingest_config.ingest_type)
    logger.info('Beginning load_and_push_jira_to_s3')
    if not ingest_config.save_locally and not ingest_config.upload_to_s3:
        logger.error(
            f'Configuration error! Ingestion configuration must have either save_locally or upload_to_s3 set to True!'
        )
        raise Exception(
            'Save Locally and Upload to S3 are both set to False! Set one to true or no data will be saved!'
        )

    jira_config = ingest_config.jira_config
    if jira_config.skip_issues and jira_config.only_issues:
        logger.warning(f"only_issues and skip_issues are both True, so all tasks will be skipped. ")
        return False

    logging_helper.send_to_agent_log_file(f"Feature flags: {jira_config.feature_flags}")

    logger.info("Using local version of ingest")

    #######################################################################
    # SET UP JIRA CONNECTIONS (Basic and Potentially Atlassian Connect)
    #######################################################################
    jira_basic_connection = get_jira_connection(
        config=jira_config, auth_method=JiraAuthMethod.BasicAuth
    )

    jira_atlas_connect_connection = (
        get_jira_connection(config=jira_config, auth_method=JiraAuthMethod.AtlassianConnect)
        if JiraAuthMethod.AtlassianConnect in jira_config.available_auth_methods
        else None
    )
    # There is an ongoing effort to cut all things over to Atlassian Connect only,
    # but it is a piecewise migration for now.
    # OJ-29745
    jira_connect_or_fallback_connection = jira_basic_connection

    if jira_config.feature_flags.get("lusca-auth-always-use-connect-for-atlassian-apis-Q423"):
        logging_helper.send_to_agent_log_file("Will use connect for most API calls")
        jira_connect_or_fallback_connection = jira_atlas_connect_connection
    else:
        logging_helper.send_to_agent_log_file("Will use basic auth for most API calls")

    #######################################################################
    # Init IO Helper
    #######################################################################
    ingest_io_helper = IngestIOHelper(ingest_config=ingest_config)

    S3_OR_JELLYFISH_LOG_STATEMENT = (
        's3' if ingest_config.ingest_type == IngestionType.DIRECT_CONNECT else 'jellyfish'
    )
    if ingest_config.save_locally:
        logger.info(f"Data will be saved locally to: {ingest_io_helper.local_file_path}")
    else:
        logger.info("Data will not be saved locally")
    if ingest_config.upload_to_s3:
        logger.info(f"Data will be submitted to {S3_OR_JELLYFISH_LOG_STATEMENT}")
    else:
        logger.info(f"Data will NOT be submitted to {S3_OR_JELLYFISH_LOG_STATEMENT}")

    #######################################################################
    # Jira Projects
    #######################################################################
    projects_and_versions = download_projects_and_versions_and_components(
        jira_connection=jira_connect_or_fallback_connection,
        is_agent_run=ingest_config.ingest_type == IngestionType.AGENT,
        jellyfish_project_ids_to_keys=jira_config.jellyfish_project_ids_to_keys,
        jellyfish_issue_metadata=jira_config.jellyfish_issue_metadata,
        include_projects=jira_config.include_projects,
        exclude_projects=jira_config.exclude_projects,
        include_categories=jira_config.include_project_categories,
        exclude_categories=jira_config.exclude_project_categories,
    )

    project_ids = {proj["id"] for proj in projects_and_versions}
    ingest_io_helper.write_json_to_local_or_s3(
        object_name=JiraObject.JiraProjectsAndVersions.value,
        json_data=projects_and_versions,
        subdirectory=SubDirectory.JIRA,
        save_locally=ingest_config.save_locally,
        upload_to_s3=ingest_config.upload_to_s3,
    )

    if not jira_config.only_issues:
        #######################################################################
        # Jira Fields
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraFields.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_fields(
                jira_connect_or_fallback_connection,
                jira_config.include_fields,
                jira_config.exclude_fields,
            ),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        ######################################################################
        # Jira Users
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraUsers.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_users(
                jira_basic_connection=jira_basic_connection,  # Use BasicAuth because /users/search is not supported by Connect apps.
                jira_atlas_connect_connection=jira_atlas_connect_connection
                if jira_config.should_augment_emails
                else None,  # Use AtlasConnect for 'augment with email' subtask
                gdpr_active=jira_config.gdpr_active,
                search_users_by_letter_email_domain=jira_config.search_users_by_letter_email_domain,
                required_email_domains=jira_config.required_email_domains,
                is_email_required=jira_config.is_email_required,
            ),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Resolutions
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraResolutions.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_resolutions(jira_connect_or_fallback_connection),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Issue Types
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraIssueTypes.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_issuetypes(
                jira_connect_or_fallback_connection, project_ids=project_ids
            ),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Link Types
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraLinkTypes.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_issuelinktypes(jira_connect_or_fallback_connection),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Priorities
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraPriorities.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_priorities(jira_connect_or_fallback_connection),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Statuses
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraStatuses.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_statuses(jira_connect_or_fallback_connection),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Boards, Sprints, and Links
        #######################################################################
        boards, sprints, links = download_boards_and_sprints(
            jira_connect_or_fallback_connection,
            jira_config.download_sprints,
        )
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraBoards.value,
            subdirectory=SubDirectory.JIRA,
            json_data=boards,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraSprints.value,
            subdirectory=SubDirectory.JIRA,
            json_data=sprints,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraBoardSprintLinks.value,
            subdirectory=SubDirectory.JIRA,
            json_data=links,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

    if not jira_config.skip_issues:
        # For Jira Issue operations we need to determine the batch size that
        # the JIRA provider will limit us to
        jira_issues_batch_size = get_jira_search_batch_size(
            jira_connection=jira_connect_or_fallback_connection,
            optimistic_batch_size=jira_config.issue_batch_size,
        )

        # check if this company is on new ingest path, default to False
        if jira_config.feature_flags.get(Constants.NEW_SYNC_FF_NAME, False):
            all_downloaded_issue_ids, all_deleted_issue_ids = pull_issues_new_sync_path(
                jira_connect_or_fallback_connection,
                jira_config,
                projects_and_versions,
                jira_issues_batch_size,
                ingest_io_helper,
                ingest_config,
            )
        else:
            all_downloaded_issue_ids, all_deleted_issue_ids = pull_issues_old_metadata_path(
                jira_connect_or_fallback_connection,
                jira_issues_batch_size,
                jira_config,
                projects_and_versions,
                ingest_io_helper,
                ingest_config,
            )

        logger.info(f'{len(all_deleted_issue_ids)} issues have been detected as being deleted')
        # Write issues that got deleted
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraIssuesIdsDeleted.value,
            subdirectory=SubDirectory.JIRA,
            json_data=list(all_deleted_issue_ids),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Work Logs
        #######################################################################
        if jira_config.download_worklogs:
            ingest_io_helper.write_json_to_local_or_s3(
                object_name=JiraObject.JiraWorklogs.value,
                subdirectory=SubDirectory.JIRA,
                json_data=download_worklogs(
                    jira_connect_or_fallback_connection,
                    all_downloaded_issue_ids,
                    jira_config.work_logs_pull_from,
                ),
                save_locally=ingest_config.save_locally,
                upload_to_s3=ingest_config.upload_to_s3,
            )
    else:
        logger.info(
            f"Skipping issues and worklogs bc config.skip_issues is {jira_config.skip_issues}"
        )

    if ingest_config.save_locally:
        logger.info(f"Data has been saved locally to: {ingest_io_helper.local_file_path}")
    else:
        logger.info(
            f"Data has not been saved locally, because save_locally was set to false in the ingest config!"
        )

    if ingest_config.upload_to_s3:
        logger.info(f"Data has been submitted to {S3_OR_JELLYFISH_LOG_STATEMENT}")
    else:
        logger.info(f"Data was not submitted to {S3_OR_JELLYFISH_LOG_STATEMENT}")

    return True



import dataclasses
import gzip
import json
import logging
import os
import time
import traceback
from datetime import datetime
from pathlib import Path
from typing import Optional

import requests

from jf_ingest import logging_helper
from jf_ingest.config import IngestionConfig, IngestionType
from jf_ingest.utils import retry_session

logger = logging.getLogger(__name__)


class SubDirectory:
    JIRA = "jira"
    GIT = "git"


class StrDefaultEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        elif type(o) == datetime:
            return o.isoformat()
        return str(o)


class IngestIOHelper:
    def __init__(self, ingest_config: IngestionConfig):
        self.ingest_config = ingest_config
        # EVERYTHING in this file path will (potentially) be uploaded to S3!
        # DO NOT put any creds file in this path!!!!
        self.local_file_path = ingest_config.local_file_path

        if not os.path.exists(self.local_file_path):
            os.makedirs(self.local_file_path)

    def _get_file_name(self, object_name: str, batch_number: Optional[int] = 0):
        return f'{object_name}{batch_number if batch_number else ""}.json'

    def _serialize_json_to_bytes(
        self, json_data: dict | list[dict], indent_level: int = 2
    ) -> bytes:
        return json.dumps(json_data, indent=indent_level, cls=StrDefaultEncoder).encode("utf-8")

    def _write_file(self, json_data: dict | list[dict], full_file_path: str):
        """
        Writes json file data to local file system at filepath. Called by write_json_data_to_local.
        :param json_data:
        :param full_file_path:
        :return: nothing
        """
        with open(full_file_path, "wb") as f:
            f.write(self._serialize_json_to_bytes(json_data=json_data))
            logger.debug(f"File: {full_file_path}, Size: {round(f.tell() / 1000000, 1)}MB")

    def write_json_to_local_or_s3(
        self,
        object_name: str,
        json_data: dict | list[dict],
        subdirectory: SubDirectory,
        save_locally: bool,
        upload_to_s3: bool,
        git_instance_key: str = None,
        batch_number: Optional[int] = 0,
    ) -> None:
        """Helper function for writing data locally or pushing data to S3.
        NOTE: When writing git data, we require an instance key. This is because
        we can have multiple git instances, so we rely on a uniqueness key to keep
        things separated. The key is generated by the GithubInstance object in Jellyfish

        Args:
            object_name (str): The object string we are writing
            json_data (dict | list[dict]): The 'raw' data
            subdirectory (SubDirectory): The directory we are pushing to (JIRA or GIT)
            save_locally (bool): If true, save locally
            upload_to_s3 (bool): If true, push to S3 (remote)
            git_instance_key (str): When provided, adds in an additional suffix to the subdirectory. Required for Git!
            batch_number (Optional[int], optional): Used for breaking up very large files, like JiraIssues. Defaults to 0.
        """
        if subdirectory == SubDirectory.GIT:
            if git_instance_key:
                subdirectory_str = f'{subdirectory}_{git_instance_key}'
            else:
                raise Exception(
                    f'When writing Git Data with this function, the file_prefix argument is required!'
                )
        else:
            subdirectory_str = str(subdirectory)

        if save_locally:
            self.write_json_data_to_local(
                object_name=object_name,
                json_data=json_data,
                subdirectory=subdirectory_str,
                batch_number=batch_number,
            )
        if upload_to_s3:
            self.write_json_data_to_s3(
                object_name=object_name,
                json_data=json_data,
                subdirectory=subdirectory_str,
                batch_number=batch_number,
            )

    def write_json_data_to_local(
        self,
        object_name: str,
        json_data: dict | list[dict],
        subdirectory: str,
        batch_number: Optional[int] = 0,
    ) -> None:
        """
        Writes json data to local file system in proper subdirectory (depending on jira or git)
        :param object_name: from JiraObject or GitObject enum, eg "jira_boards", "git_prs", etc
        :param json_data: data returned from ingestion
        :param subdirectory: directory under `<output_dir>/<timestamp>/` to write to, eg "jira" or "git_EXAMPLEKEY"
        :param batch_number: if this is a batched file, the batch number
        :return: None
        """
        file_name = self._get_file_name(object_name=object_name, batch_number=batch_number)
        full_file_path = f"{self.local_file_path}/{subdirectory}/{file_name}"
        logger.debug(f"Attempting to save {object_name} data to {full_file_path}")
        try:
            self._write_file(json_data, full_file_path)
        except FileNotFoundError:
            Path(full_file_path).parent.mkdir(exist_ok=True, parents=True)
            self._write_file(json_data, full_file_path)
        logger.debug(f"Successfully saved {object_name} data to {full_file_path}")

    def write_json_data_to_s3(
        self,
        object_name: str,
        json_data: dict | list[dict],
        subdirectory: str,
        batch_number: Optional[int] = 0,
    ) -> None:
        """Helper function for uploading JSON Data to S3, one file at a time.
        Called at potentially the same time that we save data to the local disk.
        We want to continuously upload files to S3 and potentially not save them
        locally to reduce our disk footprint when working in Dask Tasks, which
        are limited to 20 GB as of January 2024

        Args:
            object_name (str): The object string we are writing
            json_data (dict | list[dict]): The "raw" data
            subdirectory (str): The subdirectory we are pushing to (jira/ or git_EXAMPLEKEY/)
            batch_number (Optional[int], optional): Used for breaking up very large files, like JiraIssues. Defaults to 0.
        """
        # Get signed URL for this file
        filename = f"{subdirectory}/{self._get_file_name(object_name=object_name, batch_number=batch_number)}"
        s3_path, url, fields = self.get_signed_url(filename=filename)

        encoded_json_data = self._serialize_json_to_bytes(json_data=json_data, indent_level=0)
        logging_helper.send_to_agent_log_file(f"Compressing {object_name} data...")
        zipped_json_data = gzip.compress(encoded_json_data)
        session = retry_session()
        # When pushing to S3, there's a small chance that we randomly get a 500 error.
        # It is expected that on retry this should be resolved.
        # To be safe, we've implemented an exponential back off algorithm
        max_attempts = 5
        current_attempt = 0
        while True:
            try:
                logging_helper.send_to_agent_log_file(
                    f"Attempting to upload {object_name} data to S3"
                )
                upload_resp = session.post(
                    url,
                    data=fields,
                    files={"file": (s3_path, zipped_json_data)},
                )
                upload_resp.raise_for_status()
                logging_helper.send_to_agent_log_file(
                    f"Successfully uploaded {object_name} data to S3"
                )
                break
            except Exception as e:
                # Increment attempt number
                current_attempt += 1
                logging_helper.send_to_agent_log_file(
                    f"Exception encountered when attempting to upload {object_name} data to S3. Attempt number {current_attempt} of {max_attempts} Error: {e}",
                    level=logging.ERROR,
                )
                if current_attempt >= max_attempts:
                    logger.error(f'Error attempting to upload {object_name} data to S3')
                    logging_helper.send_to_agent_log_file(
                        traceback.format_exc(), level=logging.ERROR
                    )
                    raise
                else:
                    sleep_time = current_attempt**2
                    logging_helper.send_to_agent_log_file(f'Sleeping for {sleep_time}')
                    time.sleep(sleep_time)

        return

    def get_signed_url(self, filename: str) -> None:
        """
        Gets signed url from Jellyfish API for uploading to S3.
        This hits an API that is designed to return multiple signed URLs,
        but we are using it for just one upload
        :param filename: list of filenames that will need urls
        :param timestamp: timestamp created at beginning of ingestion (used for output directory)
        :return: A dictionary in the following shape:
        {
            'signed_urls: {
                <filename>: {
                    's3_path': '<s3_path>',
                    'url': {
                        'url': '<signed_url>',
                        'fields': {
                            '<field>': '<value>' for field in (key, AWSAccessKeyId, x-amz-security-token, policy, signature)
                        }
                    }
                },
            },
        }
        """
        filename = f"{filename}.gz"
        base_url = self.ingest_config.jellyfish_api_base
        headers = {"Jellyfish-API-Token": self.ingest_config.jellyfish_api_token}
        payload = {
            # NOTE: The API handles a LIST of file names, so we need to wrap this
            # individual file in brackets
            "files": [filename],
            "ingestType": self.ingest_config.ingest_type,
        }  # the files will be gzipped before uploading

        r = requests.post(
            f"{base_url}/endpoints/ingest/signed-url?timestamp={self.ingest_config.timestamp}",
            headers=headers,
            json=payload,
        )
        r.raise_for_status()
        json_resp = r.json()["signed_urls"][filename]

        return json_resp['s3_path'], json_resp['url']['url'], json_resp['url']['fields']

    def get_signed_urls(
        self, filenames: list[str], timestamp: str, ingestion_type: IngestionType
    ) -> dict[str, dict]:
        """
        Gets signed urls from Jellyfish API for uploading to S3
        :param filenames: list of filenames that will need urls
        :param timestamp: timestamp created at beginning of ingestion (used for output directory)
        :return: the signed urls dict of shape:
        {
            'signed_urls: {
                <filename>: {
                    's3_path': '<s3_path>',
                    'url': {
                        'url': '<signed_url>',
                        'fields': {
                            '<field>': '<value>' for field in (key, AWSAccessKeyId, x-amz-security-token, policy, signature)
                        }
                    }
                },
            },
        }
        """
        base_url = self.ingest_config.jellyfish_api_base
        headers = {"Jellyfish-API-Token": self.ingest_config.jellyfish_api_token}
        payload = {
            "files": [f"{name}.gz" for name in filenames],
            "ingestType": ingestion_type,
        }  # the files will be gzipped before uploading

        r = requests.post(
            f"{base_url}/endpoints/ingest/signed-url?timestamp={timestamp}",
            headers=headers,
            json=payload,
        )
        r.raise_for_status()

        return r.json()["signed_urls"]



import json

import requests_mock

from jf_ingest.jf_jira.downloaders import download_resolutions
from tests.jf_jira.utils import _register_jira_uri, get_fixture_file_data, get_jira_mock_connection


def test_download_resolutions():
    """
    Resolutions is a simple function, so we just need a simple test
    """
    with requests_mock.Mocker() as m:
        resolution_data_str = get_fixture_file_data("api_responses/resolution.json")
        _register_jira_uri(m, endpoint="resolution", return_value=resolution_data_str)

        resolutions = download_resolutions(get_jira_mock_connection())

        assert type(resolutions) == list
        assert resolutions == json.loads(resolution_data_str)



import datetime
import json
import os
from contextlib import contextmanager
from unittest.mock import patch

import pytest
import pytz
import requests_mock

from jf_ingest.jf_jira.downloaders import download_projects_and_versions_and_components
from tests.jf_jira.utils import (
    _register_jira_uri,
    _register_jira_uri_with_file,
    get_jellyfish_issue_metadata,
    get_jira_mock_connection,
)


@contextmanager
def _mock_for_get_jira_projects_and_versions():
    with requests_mock.Mocker() as _requests_mock:
        _register_jira_uri_with_file(_requests_mock, "project", "api_responses/project.json")
        _register_jira_uri_with_file(
            _requests_mock,
            "project/OJ/versions",
            "api_responses/project_versions_OJ.json",
        )
        _register_jira_uri_with_file(
            _requests_mock,
            "project/JFR/versions",
            "api_responses/project_versions_JFR.json",
        )
        _register_jira_uri_with_file(
            _requests_mock,
            "project/OJ/components",
            "api_responses/project_components_OJ.json",
        )
        _register_jira_uri(_requests_mock, "project/JFR/components", "[]")
        yield


def test_get_jira_projects_and_versions_full_test_agent_run():
    # test basic project exclusion/inclusion
    # Test when all projects are accessible
    with _mock_for_get_jira_projects_and_versions():
        projects = download_projects_and_versions_and_components(
            jira_connection=get_jira_mock_connection(),
            is_agent_run=True,
            jellyfish_project_ids_to_keys={},
            jellyfish_issue_metadata=get_jellyfish_issue_metadata(),
            include_projects=[],
            exclude_projects=[],
            include_categories=[],
            exclude_categories=[],
        )

        with open(
            f"{os.path.dirname(__file__)}/fixtures/results/project_version_component.json",
            "r",
        ) as f:
            result_project_version_component_data = f.read()
            assert projects == json.loads(result_project_version_component_data)


def test_get_jira_projects_and_versions_filtering_test_agent_run():
    # test basic project exclusion/inclusion
    # Test when all projects are accessible
    with _mock_for_get_jira_projects_and_versions():
        # Test Include Filter
        projects = download_projects_and_versions_and_components(
            jira_connection=get_jira_mock_connection(),
            is_agent_run=True,
            jellyfish_project_ids_to_keys={},
            jellyfish_issue_metadata=get_jellyfish_issue_metadata(),
            include_projects=["OJ"],
            exclude_projects=[],
            include_categories=[],
            exclude_categories=[],
        )

        assert "JFR" not in [proj["key"] for proj in projects]
        assert "OJ" in [proj["key"] for proj in projects]

        # Test Exclude Filter
        projects = download_projects_and_versions_and_components(
            jira_connection=get_jira_mock_connection(),
            is_agent_run=True,
            jellyfish_project_ids_to_keys={},
            jellyfish_issue_metadata=get_jellyfish_issue_metadata(),
            include_projects=[],
            exclude_projects=["OJ"],
            include_categories=[],
            exclude_categories=[],
        )

        assert "JFR" in [proj["key"] for proj in projects]
        assert "OJ" not in [proj["key"] for proj in projects]


# If we are NOT in an agent run (is_agent_run=False), than we should do no filtering!
def test_get_jira_projects_and_versions_filtering_test_not_agent_run():
    # test basic project exclusion/inclusion
    # Test when all projects are accessible
    with _mock_for_get_jira_projects_and_versions():
        # Test Include Filter
        projects = download_projects_and_versions_and_components(
            jira_connection=get_jira_mock_connection(),
            is_agent_run=False,
            jellyfish_project_ids_to_keys={},
            jellyfish_issue_metadata=get_jellyfish_issue_metadata(),
            include_projects=["OJ"],
            exclude_projects=[],
            include_categories=[],
            exclude_categories=[],
        )

        assert "JFR" in [proj["key"] for proj in projects]
        assert "OJ" in [proj["key"] for proj in projects]

        # Test Exclude Filter
        projects = download_projects_and_versions_and_components(
            jira_connection=get_jira_mock_connection(),
            is_agent_run=False,
            jellyfish_project_ids_to_keys={},
            jellyfish_issue_metadata=get_jellyfish_issue_metadata(),
            include_projects=[],
            exclude_projects=["OJ"],
            include_categories=[],
            exclude_categories=[],
        )

        assert "JFR" in [proj["key"] for proj in projects]
        assert "OJ" in [proj["key"] for proj in projects]


def test_get_jira_projects_and_versions_rekey_test():
    with _mock_for_get_jira_projects_and_versions():
        earliest_date = pytz.utc.localize(datetime.datetime.min)

        jellyfish_issue_metadata = get_jellyfish_issue_metadata()

        # BASELINE: TEST THAT WE DO NOT AFFECT ISSUE META UPDATED DATES AT ALL
        download_projects_and_versions_and_components(
            jira_connection=get_jira_mock_connection(),
            is_agent_run=False,
            jellyfish_project_ids_to_keys={},
            jellyfish_issue_metadata=jellyfish_issue_metadata,
            include_projects=[],
            exclude_projects=[],
            include_categories=[],
            exclude_categories=[],
        )

        oj_issue_metadata = [
            metadata for metadata in jellyfish_issue_metadata if metadata.project_id == "10000"
        ]

        assert not all(
            issue_metadata.updated == earliest_date for issue_metadata in oj_issue_metadata
        )

        # TEST THAT WE DO AFFECT ISSUE META UPDATED DATES !
        download_projects_and_versions_and_components(
            jira_connection=get_jira_mock_connection(),
            is_agent_run=False,
            jellyfish_project_ids_to_keys={"10000": "OJ-NEW"},
            jellyfish_issue_metadata=jellyfish_issue_metadata,
            include_projects=[],
            exclude_projects=[],
            include_categories=[],
            exclude_categories=[],
        )

        oj_issue_metadata = [
            metadata for metadata in jellyfish_issue_metadata if metadata.project_id == "10000"
        ]

        assert all(issue_metadata.updated == earliest_date for issue_metadata in oj_issue_metadata)



import json

import requests_mock

from jf_ingest.jf_jira.downloaders import download_statuses
from tests.jf_jira.utils import _register_jira_uri, get_fixture_file_data, get_jira_mock_connection


def test_download_statuses():
    """
    Resolutions is a simple function, so we just need a simple test
    """
    with requests_mock.Mocker() as m:
        status_data_str = get_fixture_file_data("api_responses/status.json")
        _register_jira_uri(m, endpoint="status", return_value=status_data_str)

        statuses = download_statuses(get_jira_mock_connection())

        assert type(statuses) == list
        assert [status["raw_json"] for status in statuses] == json.loads(status_data_str)
        for status_json in statuses:
            assert status_json["status_id"] == status_json["raw_json"]["id"]



from abc import ABC
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import TypeVar

IManifest = TypeVar('IManifest', bound='ManifestBase')


class MalformedJSONRepresentation(Exception):
    pass


class ManifestSource(Enum):
    # 'Our' representation of a clients Data
    # i.e. what's in our SQL database
    local = 'LOCAL'
    # 'Their' data. I.e. data in their Github,
    # Gitlab, BBCloud, or Jira instances
    remote = 'REMOTE'
    # The delta between any two manifests
    delta = 'DELTA'


# Currently used to codify the different types
# of manifests accessible via pulling from S3
# This field IS NOT used in the actual manifest class!
# The actual manifest class uses only str types for
# determining its type; this is to leave the typing
# as generalized as possible
class ManifestType(Enum):
    git = 'GitDataManifest'
    jira = 'JiraDataManifest'


# Separating the model from the adapter behavior as a base class to make the data more transportable and reusable
@dataclass
class ManifestBase(ABC):
    company: str
    # Where the data came from (local, remote, or delta)
    data_source: ManifestSource
    date: datetime = field(init=False)
    manifest_type: str = field(init=False)
    _module_name: str = field(init=False)



import re


class NameRedactor:
    """This is a helper class for cleaning up sensitive data that gets submitted
    by the agent. It is currently only used by the Git ingest functionality.

    The boolean fields related to this logic are found on the GitConfig. They are:
        GitConfig.git_strip_text_content,
        GitConfig.git_redact_names_and_urls,
    """

    def __init__(self, preserve_names=None):
        self.redacted_names = {}
        self.seq = 0
        self.preserve_names = preserve_names or []

    def redact_name(self, name):
        if not name or name in self.preserve_names:
            return name

        redacted_name = self.redacted_names.get(name)
        if not redacted_name:
            redacted_name = f'redacted-{self.seq:04}'
            self.seq += 1
            self.redacted_names[name] = redacted_name
        return redacted_name


def sanitize_text(text, strip_text_content):
    # NOTE: This module is used only by git, but we need to clean up Jira
    # keys out of commit messages. That's what this regex is for
    JIRA_KEY_REGEX = re.compile(r'([a-z0-9]+)[-|_|/| ]?(\d+)', re.IGNORECASE)

    if not text or not strip_text_content:
        return text

    return (' ').join(
        {f'{m[0].upper().strip()}-{m[1].upper().strip()}' for m in JIRA_KEY_REGEX.findall(text)}
    )



import json
import os

import requests_mock
from jira import JIRA

from jf_ingest.config import JiraAuthConfig
from jf_ingest.jf_jira.auth import get_jira_connection
from jf_ingest.jf_jira.downloaders import IssueMetadata

_MOCK_REST_BASE_URL = "https://test-co.atlassian.net/rest/api/2"
_MOCK_AGILE_BASE_URL = "https://test-co.atlassian.net/rest/agile/1.0"


def get_jira_mock_connection(mocker=None) -> JIRA:
    """Helper function for getting a spoofed JIRA connection

    Returns:
        JIRA: A Mocked JIRA connection object
    """

    _MOCK_SERVER_INFO_RESP = (
        '{"baseUrl":"https://test-co.atlassian.net","version":"1001.0.0-SNAPSHOT",'
        '"versionNumbers":[1001,0,0],"deploymentType":"Cloud","buildNumber":100218,'
        '"buildDate":"2023-03-16T08:21:48.000-0400","serverTime":"2023-03-17T16:32:45.255-0400",'
        '"scmInfo":"9999999999999999999999999999999999999999","serverTitle":"JIRA",'
        '"defaultLocale":{"locale":"en_US"}} '
    )

    auth_config = JiraAuthConfig(
        url="https://test-co.atlassian.net/",
        personal_access_token="asdf",
        company_slug="test_co",
    )

    if not mocker:
        with requests_mock.Mocker() as m:
            _register_jira_uri(m, "serverInfo", f"{_MOCK_SERVER_INFO_RESP}")
            _register_jira_uri_with_file(m, "field", "api_responses/fields.json")
            jira_conn = get_jira_connection(config=auth_config, max_retries=1)
    else:
        _register_jira_uri(mocker, "serverInfo", f"{_MOCK_SERVER_INFO_RESP}")
        _register_jira_uri_with_file(mocker, "field", "api_responses/fields.json")
        jira_conn = get_jira_connection(config=auth_config, max_retries=1)

    return jira_conn


def get_fixture_file_data(fixture_path: str) -> str:
    """Helper function to get file data from a fixture

    Usage (if calling from tests/jf_jira/test_jira_resolutions.py):
    fixture_path='api_response/resolutions.json'

    Args:
        fixture_path (str): A path to a fixture file. It prepends the current relative path.

    Returns:
        str: The file data as a string
    """
    with open(f"{os.path.dirname(__file__)}/fixtures/{fixture_path}", "r") as f:
        return f.read()


def get_jellyfish_issue_metadata() -> dict[str, IssueMetadata]:
    """Gets some hardcoded jellyfish metadata

    Returns:
        dict[str, dict]: Hardcoded jellyfish issue data from tests/jf_jira/fixtures/jellyfish_issue_metadata.json
    """
    with open(f"{os.path.dirname(__file__)}/fixtures/jellyfish_issue_metadata.json", "r") as f:
        return [IssueMetadata(**metadata_dict) for metadata_dict in json.loads(f.read())]


def _register_jira_uri_with_file(
    mock: requests_mock.Mocker, endpoint: str, fixture_path: str
) -> None:
    """Registers a Jira API endpoint to respond with the data from a fixture file

    Args:
        mock (requests_mock.Mocker): A requests_mock.Mocker() object
        endpoint (str): The endpoint we want to spoof
        fixture_path (str): The file of data we want it to return
    """
    _register_jira_uri(mock, endpoint, get_fixture_file_data(fixture_path=fixture_path))


def _register_jira_uri(
    mock: requests_mock.Mocker,
    endpoint: str,
    return_value: str,
    HTTP_ACTION: str = "GET",
    use_agile_endpoint: bool = False,
    status_code: int = 200,
) -> None:
    """Helper function used to registering mock results for testing against our mock JIRA API. Works by providing
    a request_mock.Mocker instance, what endpoint you want to spoof, and the string that you want that spoofed endpoint
    to return

    Args:
        mock (requests_mock.Mocker): Mocker object
        endpoint (str): Endpoint to append to the end of the base URL ("https://test-co.atlassian.net/rest/api/2", unless use_agile_endpoint is True)
        return_value (str): The raw str to return from the API endpoint
        use_agile_endpoint (bool, optional): IF TRUE, switch the default base URL from the /rest/api/2 endpoint to the agile endpoint. MUST BE TRUE FOR SPRINT AND BOARD API ENDPOINTS. Defaults to False.
        HTTP_ACTION (str, optional): If we are mocking a GET or POST or PUT HTTP command. Defaults to "GET".
    """

    _MOCK_REST_BASE_URL = "https://test-co.atlassian.net/rest/api/2"

    mock_base_url = _MOCK_REST_BASE_URL if not use_agile_endpoint else _MOCK_AGILE_BASE_URL
    mock.register_uri(
        HTTP_ACTION,
        f"{mock_base_url}/{endpoint}",
        text=return_value,
        status_code=status_code,
    )
